#!/usr/bin/env python
#
# addroihistogram.py - The AddROIHistogramAction class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`AddROIHistogramAction` class, an
action used by the :class:`.HistogramPanel`.
"""


import          wx
import numpy as np

import fsl.data.image                          as fslimage
import fsleyes.views.histogrampanel            as histogrampanel
import fsleyes.strings                         as strings
import fsleyes.plotting.dataseries             as dataseries
import fsleyes.plotting.histogramseries        as histogramseries
import fsleyes.plugins.tools.addmaskdataseries as addmaskdataseries
import fsleyes.actions.base                    as base



class AddROIHistogramAction(base.Action):
    """The ``AddROIHistogramAction`` class is used by the
    :class:`.HistogramPanel`.

    It performs a very similar task to the :class:`.AddMaskDataSeriesAction` -
    the user selects a binary mask, the data within the base image is extracted
    for that mask, and the histogram of that data is added to the plot.
    """


    @staticmethod
    def supportedViews():
        """The ``AddROIHistogramAction`` is restricted for use with the
        :class:`.HistogramPanel`.
        """
        return [histogrampanel.HistogramPanel]


    def __init__(self, overlayList, displayCtx, plotPanel):
        """Create an ``AddROIHistogramAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg plotPanel:   The :class:`.HistogramPanel`.
        """

        base.Action.__init__(
            self, overlayList, displayCtx, self.__addROIHistogram)

        self.__plotPanel  = plotPanel
        self.__roiOptions = []

        overlayList.addListener('overlays',
                                self.name,
                                self.__overlayListChanged)
        displayCtx .addListener('selectedOverlay',
                                self.name,
                                self.__overlayListChanged)

        self.__overlayListChanged()


    def destroy(self):
        """Must be called when this ``AddROIHistogramAction`` is no
        longer in use.
        """
        if self.destroyed:
            return
        self.overlayList.removeListener('overlays',        self.name)
        self.displayCtx .removeListener('selectedOverlay', self.name)
        self.__plotPanel  = None
        self.__roiOptions = None
        base.Action.destroy(self)


    def __overlayListChanged(self, *a):
        """Called when the :class:`.OverlayList` or the
        :attr:`.DisplayContext.selectedOverlay` changes. Updates a list of
        valid mask images for the currently selected overlay.
        """

        overlay = self.displayCtx.getSelectedOverlay()

        if (len(self.overlayList) == 0 or
           (not isinstance(overlay, fslimage.Image))):
            self.enabled = False
            return

        self.__roiOptions = [o for o in self.overlayList if
                             isinstance(o, fslimage.Image) and
                             o is not overlay              and
                             o.sameSpace(overlay)]

        self.enabled = len(self.__roiOptions)  > 0



    def __addROIHistogram(self):
        """Prompts the user to select an ROI mask, calculates the histogram
        of that mask on the currently selected overlay, and adds the result
        to the ``HistogramPanel``.
        """

        overlay    = self.displayCtx.getSelectedOverlay()
        opts       = self.displayCtx.getOpts(overlay)
        roiOptions = self.__roiOptions

        frame   = wx.GetApp().GetTopWindow()
        msg     = strings.messages[self, 'selectMask'].format(overlay.name)
        title   = strings.titles[  self, 'selectMask'].format(overlay.name)

        dlg = addmaskdataseries.MaskDialog(
            frame,
            [o.name for o in roiOptions],
            title=title,
            message=msg,
            checkbox=False)

        if dlg.ShowModal() != wx.ID_OK:
            return

        maskimg = roiOptions[dlg.GetChoice()]
        mask    = maskimg[:] > 0

        if overlay.ndim > 3: data = overlay[opts.index()][mask]
        else:                data = overlay[mask]

        count           = self.__plotPanel.histType == 'count'
        drange          = (np.nanmin(data), np.nanmax(data))
        nbins           = histogramseries.autoBin(data, drange)
        xdata, ydata, _ = histogramseries.histogram(data,
                                                    nbins,
                                                    drange,
                                                    drange,
                                                    includeOutliers=False,
                                                    count=count)

        ds           = dataseries.DataSeries(overlay,
                                             self.overlayList,
                                             self.displayCtx,
                                             self.__plotPanel)
        ds.colour    = self.__plotPanel.getOverlayPlotColour(overlay)
        ds.lineStyle = self.__plotPanel.getOverlayPlotStyle(overlay)
        ds.lineWidth = 2
        ds.alpha     = 1
        ds.label     = '{} [mask: {}]'.format(overlay.name, maskimg.name)

        # We have to run the data through
        # prepareDataSeries to preprocess
        # (e.g. smooth) it
        ds.setData(xdata, ydata)
        ds.setData(*self.__plotPanel.prepareDataSeries(ds))

        self.__plotPanel.canvas.dataSeries.append(ds)



#!/usr/bin/env python
#
# tensoropts.py - The TensorOpts class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`TensorOpts` class, which defines
options for displaying :class:`.GLTensor` instances.
"""


import fsleyes_props as props
from . import           vectoropts


class TensorOpts(vectoropts.NiftiVectorOpts):
    """The ``TensorOpts`` class defines options for displaying :class:`.GLTensor`
    instances.
    """


    lighting = props.Boolean(default=True)
    """Enables a basic lighting model on the tensor ellipsoids. """


    tensorResolution = props.Int(minval=4, maxval=20, default=10)
    """Tensor ellipsoid resolution - this value controls the number of vertices
    used to represent each tensor. It is ultimately passed to the
    :func:`.routines.unitSphere` function.
    """


    tensorScale = props.Percentage(minval=50, maxval=600, default=100)
    """Scaling factor - by default, the tensor ellipsoids are scaled such that
    the biggest tensor (as defined by the first principal eigenvalue) fits
    inside a voxel. This parameter can be used to adjust this scale.
    """


    def __init__(self, *args, **kwargs):
        """Create a ``TensorOpts`` instance. All arguments are passed through
        to :meth:`.VectorOpts.__init__`.
        """

        vectoropts.NiftiVectorOpts.__init__(self, *args, **kwargs)



#!/usr/bin/env python
#
# gllinevector.py - The GLLineVector class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`GLLineVector` class, for displaying 3D
vector :class:`.Image` overlays in line mode.

The :class:`.GLLineVertices` class is also defined in this module, and is used
when running in OpenGL 1.4. See the :mod:`.gl14.gllinevector_funcs` and
:mod:`.gl21.gllinevector_funcs` modules for more details.
"""

import logging

import numpy                as np
import fsl.data.dtifit      as dtifit
import fsl.transform.affine as affine
import fsleyes.gl           as fslgl
import fsleyes.gl.routines  as glroutines
import fsleyes.gl.glvector  as glvector


log = logging.getLogger(__name__)


class GLLineVector(glvector.GLVector):
    """The ``GLLineVector`` class encapsulates the logic required to render an
    :class:`.Image` instance of shape ``x*y*z*3``, or type
    ``NIFTI_TYPE_RGB24``, as a vector image, where the vector at each voxel is
    drawn as a line, and coloured in the same way that voxels in the
    :class:`.GLRGBVector` are coloured.  The ``GLLineVector`` class assumes
    that the :class:`.Display` instance associated with the ``Image`` overlay
    holds a reference to a :class:`.LineVectorOpts` instance, which contains
    ``GLLineVector``-specific display settings.  The ``GLLineVector`` class is
    a sub-class of the :class:`.GLVector` class, and uses the functionality
    provided by ``GLVector``.


    In a similar manner to the :class:`.GLRGBVector`, the ``GLLineVector`` uses
    two OpenGL version-specific modules, the :mod:`.gl14.gllinevector_funcs`
    and :mod:`.gl21.gllinevector_funcs` modules. It is assumed that these
    modules define the same functions that are defined by the
    :class:`.GLRGBVector` version specific modules.


    A ``GLLineVector`` instance is rendered in different ways depending upon
    the rendering environment (GL 1.4 vs GL 2.1), so most of the rendering
    functionality is implemented in the version-specific modules mentioned
    above.

    In both cases, lines are drawn as rectangles (with each rectangle made up
    of two triangles) , so that they can be drawn at arbitrary widths.
    """


    def __init__(self, image, overlayList, displayCtx, threedee):
        """Create a ``GLLineVector`` instance.

        :arg image:       An :class:`.Image` or :class:`.DTIFitTensor`
                          instance.
        :arg overlayList: The :class:`.OverlayList`
        :arg displayCtx:  The :class:`.DisplayContext` managing the scene.
        :arg threedee:    2D or 3D rendering.
        """

        # If the overlay is a DTIFitTensor, use the
        # V1 image is the vector data. Otherwise,
        # assume that the overlay is the vector image.
        if isinstance(image, dtifit.DTIFitTensor): vecImage = image.V1()
        else:                                      vecImage = image

        def prefilter(data):
            # Scale to unit length if required.
            if not self.opts.unitLength:
                return data

            data = np.copy(data)

            with np.errstate(invalid='ignore'):

                # Vector images stored as RGB24 data
                # type are assumed to map from [0, 255]
                # to [-1, 1], so cannot be normalised
                if vecImage.nvals > 1:
                    return data

                # calculate lengths
                x    = data[0, ...]
                y    = data[1, ...]
                z    = data[2, ...]
                lens = np.sqrt(x ** 2 + y ** 2 + z ** 2)

                # scale lengths to 1
                data[0, ...] = x / lens
                data[1, ...] = y / lens
                data[2, ...] = z / lens

            return data

        def prefilterRange(dmin, dmax):

            if self.opts.unitLength:
                return -1, 1
            else:
                return dmin, dmax

        glvector.GLVector.__init__(self,
                                   image,
                                   overlayList,
                                   displayCtx,
                                   threedee,
                                   prefilter=prefilter,
                                   prefilterRange=prefilterRange,
                                   vectorImage=vecImage,
                                   init=lambda: fslgl.gllinevector_funcs.init(
                                       self))

        self.opts.addListener('lineWidth',  self.name, self.notify)
        self.opts.addListener('unitLength', self.name,
                              self.__unitLengthChanged)


    def destroy(self):
        """Must be called when this ``GLLineVector`` is no longer needed.
        Removes some property listeners from the :class:`.LineVectorOpts`
        instance, calls the OpenGL version-specific ``destroy``
        function, and calls the :meth:`.GLVector.destroy` method.
        """
        self.opts.removeListener('lineWidth',  self.name)
        self.opts.removeListener('unitLength', self.name)
        fslgl.gllinevector_funcs.destroy(self)
        glvector.GLVector.destroy(self)


    def normalisedLineWidth(self, canvas):
        """Returns the :attr:`.LineVectorOpts.lineWidth`, scaled to be
        proportional to the same coordinate system that the vector values
        are defined in (assumed to be scaled voxels, e.g. the FSL coordinate
        system).
        """
        opts      = self.opts
        d2p       = opts.getTransform('display', 'pixdim')
        lineWidth = opts.lineWidth * canvas.pixelSize()[0]
        lineWidth = affine.transform([lineWidth] * 3, d2p, vector=True)[0]
        return lineWidth / 2


    def getDataResolution(self, xax, yax):
        """Overrides :meth:`.GLImageObject.getDataResolution`. Returns a pixel
        resolution suitable for rendering this ``GLLineVector``.
        """

        res       = list(glvector.GLVector.getDataResolution(self, xax, yax))
        res[xax] *= 20
        res[yax] *= 20

        return res


    def compileShaders(self):
        """Overrides :meth:`.GLVector.compileShaders`. Calls the OpenGL
        version-specific ``compileShaders`` function.
        """
        fslgl.gllinevector_funcs.compileShaders(self)


    def updateShaderState(self):
        """Overrides :meth:`.GLVector.updateShaderState`. Calls the OpenGL
        version-specific ``updateShaderState`` function.
        """
        return fslgl.gllinevector_funcs.updateShaderState(self)


    def preDraw(self):
        """Overrides :meth:`.GLVector.preDraw`. Calls the base class
        implementation, and then calls the OpenGL version-specific ``preDraw``
        function.
        """
        glvector.GLVector.preDraw(self)
        fslgl.gllinevector_funcs.preDraw(self)


    def draw2D(self, *args, **kwargs):
        """Overrides :meth:`.GLObject.draw2D`. Calls the OpenGL
        version-specific ``draw2D`` function.
        """
        fslgl.gllinevector_funcs.draw2D(self, *args, **kwargs)


    def draw3D(self, *args, **kwargs):
        """Does nothing. """


    def drawAll(self, *args, **kwargs):
        """Overrides :meth:`.GLObject.drawAll`. Calls the OpenGL
        version-specific ``drawAll`` function.
        """
        fslgl.gllinevector_funcs.drawAll(self, *args, **kwargs)


    def postDraw(self):
        """Overrides :meth:`.GLVector.postDraw`. Calls the base class
        implementation, and then calls the OpenGL version-specific ``postDraw``
        function.
        """
        glvector.GLVector.postDraw(self)
        fslgl.gllinevector_funcs.postDraw(self)


    def __unitLengthChanged(self, *a):
        """Called when the :attr:`.LineVectorOptsunitLength` property
        changes. Refreshes the vector image texture data.
        """
        self.imageTexture.refresh()
        self.updateShaderState()


class GLLineVertices:
    """The ``GLLineVertices`` class is used when rendering a
    :class:`GLLineVector` with OpenGL 1.4. It contains logic to generate
    vertices for every vector in the vector :class:`.Image` that is being
    displayed by a ``GLLineVector`` instance.


    This class is used by the OpenGL 1.4 implementation - when using OpenGL
    2.1, the logic encoded in this class is implemented in the line vector
    vertex shader. This is because OpenGL 1.4 vertex programs (using the
    ARB_vertex_program extension) are unable to perform texture lookups,
    so cannot retrieve the vector data.


    After a ``GLLineVertices`` instance has been created, the :meth:`refresh`
    method can be used to generate line vector vertices and voxel
    coordinates for every voxel in the :class:`Image`. These vertices and
    coordinates are stored as attributes of the ``GLLineVertices`` instance.


    Later, when the line vectors from a 2D slice  of the image need to be
    displayed, the :meth:`getVertices2D` method can be used to extract the
    vertices and coordinates from the slice.


    A ``GLLineVertices`` instance is not associated with a specific
    ``GLLineVector`` instance. This is so that a single  ``GLLineVertices``
    instance can be shared between more than one ``GLLineVector``, avoiding
    the need to store multiple copies of the vertices and voxel
    coordinates. This means that a ``GLLineVector`` instance needs to be
    passed to most of the methods of a ``GLLineVertices`` instance.
    """

    def __init__(self, glvec):
        """Create a ``GLLineVertices``. Vertices are calculated for the
        given :class:`.GLLineVector` instance.

        :arg glvec: A :class:`GLLineVector` which is using this
                    ``GLLineVertices`` instance.
        """

        self.__hash = None
        self.refresh(glvec)


    def destroy(self):
        """Should be called when this ``GLLineVertices`` instance is no
        longer needed. Clears references to cached vertices/coordinates.
        """
        self.vertices  = None
        self.voxCoords = None


    def __hash__(self):
        """Returns a hash of this ``GLLineVertices`` instance. The hash value
        is calculated and cached on every call to :meth:`refresh`, using the
        :meth:`calculateHash` method.  This method returns that cached value.
        """
        return self.__hash


    def calculateHash(self, glvec):
        """Calculates and returns a hash value that can be used to determine
        whether the vertices of this this ``GLLineVertices`` instance need to
        be recalculated. The hash value is based on some properties of the
        :class:`.LineVectorOpts` instance, associated with the given
        :class:`.GLLineVector`.

        For a ``GLLineVertices`` instance called ``verts``, if the following
        test::

            hash(verts) != verts.calculateHash(glvec)

        evaluates to ``False``, the vertices need to be refreshed (via a
        call to :meth:`refresh`).
        """
        opts = glvec.opts
        return (hash(opts.transform)  ^
                hash(opts.orientFlip) ^
                hash(opts.directed)   ^
                hash(opts.unitLength) ^
                hash(opts.lengthScale))


    def refresh(self, glvec):
        """(Re-)calculates the vertices  of this ``GLLineVertices`` instance.

        For each voxel, in the :class:`.Image` overlay being displayed by the
        :class:`GLLineVector` associated with this ``GLLineVertices``
        instance, two vertices are generated, which define a line that
        represents the vector at the voxel.

        The vertices are stored as a :math:`X\\times Y\\times Z\\times
        2\\times 3` ``numpy`` array, as an attribute of this instance,
        called ``vertices``.
        """

        opts  = glvec.opts
        image = glvec.vectorImage
        data  = image.data
        shape = image.shape

        # Pull out the xyz components of the
        # vectors, and calculate vector lengths.

        # The image may either
        # have shape (X, Y, Z, 3)
        if image.nvals == 1:
            vertices = np.array(data, dtype=np.float32)
        # Or (we assume) a RGB
        # structured array
        else:
            vertices         = np.zeros(list(data.shape) + [3],
                                        dtype=np.float32)
            vertices[..., 0] = (data['R'].astype(np.float32) / 127.5) - 1
            vertices[..., 1] = (data['G'].astype(np.float32) / 127.5) - 1
            vertices[..., 2] = (data['B'].astype(np.float32) / 127.5) - 1

        x    = vertices[..., 0]
        y    = vertices[..., 1]
        z    = vertices[..., 2]
        lens = np.sqrt(x ** 2 + y ** 2 + z ** 2)

        # Flip vectors about the x axis if necessary
        if opts.orientFlip:
            x = -x

        if opts.unitLength:

            # scale the vector lengths to 0.5
            with np.errstate(invalid='ignore'):
                vertices[..., 0] = 0.5 * x / lens
                vertices[..., 1] = 0.5 * y / lens
                vertices[..., 2] = 0.5 * z / lens

            # Scale the vector data by the minimum
            # voxel length, so it is a unit vector
            # within real world space. We're assuming
            # here that the vectors are defined in
            # mm (e.g. the FSL coordinate system).
            vertices /= (image.pixdim[:3] / min(image.pixdim[:3]))

        # Scale the vectors by the length scaling factor
        vertices *= opts.lengthScale / 100.0

        # Duplicate vector data so that each
        # vector is represented by two vertices,
        # representing a line through the origin.
        # Or, if displaying directed vectors,
        # add an origin point for each vector.
        if opts.directed:
            origins  = np.zeros(vertices.shape, dtype=np.float32)
            vertices = np.concatenate((origins, vertices), axis=3)
        else:
            vertices = np.concatenate((-vertices, vertices), axis=3)

        vertices = vertices.reshape((shape[0],
                                     shape[1],
                                     shape[2],
                                     2,
                                     3))

        self.vertices = vertices
        self.__hash   = self.calculateHash(glvec)


    def getVertices2D(self, glvec, canvas, zpos, axes, bbox=None):
        """Extracts and returns a slice of line vertices, and the associated
        voxel coordinates, which are in a plane located at the given Z
        position (in display coordinates).

        The line vertices are transformed into rectangular polygons, suitable
        for being drawn with the ``GL_TRIANGLES`` primitive.

        This method assumes that the :meth:`refresh` method has already been
        called.
        """

        opts   = glvec.opts
        image  = glvec.image
        shape  = image.shape[:3]
        zax    = axes[2]

        vertices  = self.vertices
        voxCoords = glvec.generateVoxelCoordinates2D(zpos, axes, bbox)

        # Turn the voxel coordinates into
        # indices suitable for looking up
        # the corresponding vertices
        coords = np.array(np.floor(voxCoords + 0.5), dtype=np.int32)

        # remove any out-of-bounds voxel coordinates
        shape     = np.array(vertices.shape[:3])
        inBounds  = ((coords >= [0, 0, 0]) & (coords < shape)).all(1)
        coords    = coords[   inBounds, :].T
        voxCoords = voxCoords[inBounds, :]

        # pull out the vertex data
        # corresponding to the voxels
        vertices = vertices[coords[0], coords[1], coords[2], :, :]
        vertices = vertices.reshape(-1, 3)

        # Convert line segments into rectangles so
        # we can draw lines at arbitrary widths.
        lineWidth         = glvec.normalisedLineWidth(canvas)
        vertices, indices = glroutines.lineAsPolygon(vertices,
                                                     lineWidth,
                                                     zax,
                                                     indices=True)

        if not vertices.flags['C_CONTIGUOUS']:
            vertices = np.ascontiguousarray(vertices)

        # We are drawing each line with four vertices
        # (rectangle made of two triangles). So we
        # need to repeat voxel coordinates for each
        # vertex.
        voxCoords = voxCoords.repeat(repeats=4, axis=0)

        return vertices, indices, voxCoords



#!/usr/bin/env python
#
# controlpanel.py - The ControlPanel class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`ControlPanel`, :class:`ControlToolBar`,
and :class:`SettingsPanel` classes, base-classes for all FSLeyes controls.

See the :mod:`fsleyes` package documentation for an overview of FSLeyes views
and controls.

The :class:`SettingsPanel` is a convenience
class for certain *FSLeyes* control panels (see the ::mod:`fsleyes`
documentation).
"""


import                   wx
import wx.lib.agw.aui as wxaui

import fsl.utils.deprecated       as deprecated
import fsleyes_widgets.widgetlist as widgetlist
import fsleyes.panel              as fslpanel
import fsleyes.toolbar            as fsltoolbar


class ControlMixin:
    """Mixin class for the :class:`ControlPanel` and :class:`ControlToolBar`.
    """

    @staticmethod
    def title():
        """May be overridden by sub-classes. Returns a title to be used
        in menus and window title bars.
        """
        return None


    @staticmethod
    def ignoreControl():
        """Tell FSLeyes that this control should not be considered a FSLeyes
        plugin.

        The default implementation returns ``False``, but may be overridden
        by sub-classes to control whether a menu item should be added for
        the control in the settings menu for the relevant FSLeyes view(s).

        Note that this method must be implemented on the class that is to
        be ignored - inherited implementations from base classes are not
        considered.
        """
        return False


    @staticmethod
    def supportedViews():
        """Return the views that this control supports.

        This method may be overridden by sub-classes to return a list of
        the view types that are supported by this control. For example, the
        :class:`.OrthoToolBar` control is intended to be used solely with the
        :class:`.OrthoPanel` view.

        The default implementation returns ``None``, which is interpreted
        as being compatible with all views.
        """
        return None


    @staticmethod
    def supportSubClasses():
        """May be overridden by sub-classes. If this control panel is intended
        for use with specific views, via the :meth:`supportedViews` method,
        this method specifies whether sub-classes of the :meth:`supportedViews`
        are supported (the default), or whether only the specific classes in
        :meth:`supportedViews` are supported.

        Note that this method must be implemented on the specific control
        class that wishes to use it - it is not considered if implemented in a
        base classe.
        """
        return True


    @staticmethod
    def defaultLayout():
        """Return default options for :meth:`.ViewPanel.togglePanel`.

        This method may be overridden by sub-classes to return a ``dict``
        containing keyword arguments to be used by the
        :meth:`.ViewPanel.togglePanel` method when a control panel of this
        type is added. When ``togglePanel`` is called, if any arguments are
        provided, the arguments returned by this method are not used.
        """
        return None


    @staticmethod
    def profileCls():
        """Return the custom interaction profile associated with this control.
        Must be a sub-class of :class:`.Profile`.

        When a control panel is opened, if it requires a different profile to
        the one that is active, that profile is created and activated. If any
        other controls required the previous profile, they are closed.

        Control panels which are associated with an interaction profile can
        assume that the profile has already been created by the time the
        control is created. The :class:`.Profile` instance can be retrieved
        via :meth:`.ViewPanel.currentProfile`.
        """
        return None


class ControlPanel(fslpanel.FSLeyesPanel, ControlMixin):
    """The ``ControlPanel`` is the base class for all FSLeyes controls.
    All sub-classes must call ``__init__``.
    """


    def __init__(self,
                 parent,
                 overlayList,
                 displayCtx,
                 viewPanel,
                 *args,
                 **kwargs):
        """Create a ``ControlPanel``.

        :arg parent:      Parent ``wx`` object
        :arg overlayList: The :class:`.OverlayList`
        :arg displayCtx:  The :class:`.DisplayContext` associated with
                          ``viewPanel``.
        :arg viewPanel:   The FSLeyes :class:`.ViewPanel` that owns this
                          ``ControlPanel``

        All other arguments are passed through to the ``FSLeyesPanel``.
        """

        # Prior to FSLeyes 1.0.0, the fourth argument
        # was the FSLeyesFrame. This clause is in place
        # to retain support for third-party FSLeyes
        # plugins which were written before 1.0.0
        import fsleyes.frame as fslframe
        if isinstance(viewPanel, fslframe.FSLeyesFrame):
            frame     = viewPanel
            viewPanel = None
            deprecated.warn(
                type(self).__name__, '1.0.0', '2.0.0',
                'You should pass the view panel, not the FSLeyesFrame')
        else:
            frame = viewPanel.frame

        fslpanel.FSLeyesPanel.__init__(
            self, parent, overlayList, displayCtx, frame, *args, **kwargs)
        self.__viewPanel = viewPanel


    def destroy(self):
        """Must be called when this ``ControlPanel`` is no longer needed.
        Clears references and calls the base class ``destroy`` method.
        """
        fslpanel.FSLeyesPanel.destroy(self)
        self.__viewPanel = None


    @property
    def viewPanel(self):
        """Returns a reference to the :class:`.ViewPanel` that owns this
        :class:`.ControlPanel`.
        """
        return self.__viewPanel


class ControlToolBar(fsltoolbar.FSLeyesToolBar, ControlMixin):
    """The ``ControlToolBar`` is the base class for all FSLeyes toolbars. """
    pass


class SettingsPanel(ControlPanel):
    """The ``SettingsPanel`` is a convenience class for *FSLeyes* control
    panels which use a :class:`fsleyes_widgets.WidgetList` to display a
    collection of controls for the user.  When displayed as a dialog/
    floating frame, the ``SettingsPanel`` will automatically resize itself
    to fit its contents. See the :class:`.CanvasSettingsPanel` for an example.
    """


    def __init__(self, *args, **kwargs):
        """Create a ``SettingsPanel``.  All arguments are passed to
        the :meth:`FSLeyesPanel.__init__` method.
        """

        ControlPanel.__init__(self, *args, **kwargs)

        self.__widgets = widgetlist.WidgetList(self, minHeight=24)
        self.__sizer   = wx.BoxSizer(wx.VERTICAL)

        self.SetSizer(self.__sizer)

        self.__sizer.Add(self.__widgets, flag=wx.EXPAND, proportion=1)

        self.SetMinSize((300, 80))

        self.__widgets.Bind(widgetlist.EVT_WL_CHANGE_EVENT,
                            self.__widgetListChange)


    def getWidgetList(self):
        """Returns the :class:`fsleyes_widgets.WidgetList` which should be used
        by sub-classes to display content to the user.
        """
        return self.__widgets


    def __widgetListChange(self, ev):
        """Called whenever the widget list contents change. If this panel
        is floating, its parent is autmatically resized.
        """
        if isinstance(self.GetTopLevelParent(), wxaui.AuiFloatingFrame):
            self.SetInitialSize(self.__widgets.GetBestSize())
            self.GetTopLevelParent().Fit()



#!/usr/bin/env python
#
# panel.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`FSLeyesPanel` class.


A :class:`FSLeyesPanel` object is a :class:`wx.Panel` which provides some sort
of view of a collection of overlay objects, contained within an
:class:`.OverlayList`. The :class:`FSLeyesPanel` is the base class for all GUI
panels in FSLeyes - see also the :class:`.ViewPanel` and :class:`.ControlPanel`
classes.


``FSLeyesPanel`` instances are also :class:`.ActionProvider` instances - any
actions which are specified in the class definitions may (or may not) be
exposed to the user. Furthermore, any display configuration options which
should be made available available to the user can be added as
:class:`.PropertyBase` attributes of the :class:`FSLeyesPanel` subclass.


.. note:: ``FSLeyesPanel`` instances are usually displayed within a
          :class:`.FSLeyesFrame`, but they can  be used on their own
          as well. You will need to create, or need references to,
          an :class:`.OverlayList` and a :class:`.DisplayContext`.
          For example::

              import fsleyes.overlay          as ovl
              import fsleyes.displaycontext   as dc
              import fsleyes.views.orthopanel as op

              overlayList = ovl.OverlayList()
              displayCtx  = dc.DisplayContext(overlayList)

              # the parent argument is some wx parent
              # object such as a wx.Frame or wx.Panel.
              # Pass in None as the FSLeyesFrame
              orthoPanel  = op.OrthoPanel(parent,
                                          overlayList,
                                          displayCtx,
                                          None)
"""


import logging

import              wx
import wx.siplib as sip

import fsleyes_props               as props
import fsleyes_widgets.floatspin   as floatspin
import fsleyes_widgets.floatslider as floatslider
import fsleyes_widgets.rangeslider as rangeslider

from . import                         actions
from . import                         displaycontext


log = logging.getLogger(__name__)


class _FSLeyesPanel(actions.ActionProvider, props.SyncableHasProperties):
    """The ``_FSLeyesPanel`` is the base class for the :class:`.FSLeyesPanel`
    and the :class:`.FSLeyesToolBar`.


    A ``_FSLeyesPanel`` has the following methods and properties, available for
    use by subclasses:

    .. autosummary::
       :nosignatures:

       name
       frame
       setNavOrder
       destroy
       destroyed


    .. note:: When a ``_FSLeyesPanel`` is no longer required, the
              :meth:`destroy` method **must** be called!
    """


    def __init__(self, overlayList, displayCtx, frame, kbFocus=False):
        """Create a :class:`_FSLeyesPanel`.

        :arg overlayList: A :class:`.OverlayList` instance.

        :arg displayCtx:  A :class:`.DisplayContext` instance.

        :arg frame:       The :class:`.FSLeyesFrame` that created this
                          ``_FSLeyesPanel``. May be ``None``.

        :arg kbFocus:     If ``True``, a keyboard event handler is configured
                          to intercept ``Tab`` and ``Shift+Tab`` keyboard
                          events, to shift focus between a set of child
                          widgets. The child widgets to be included in the
                          navigation can be specified with the
                          :meth:`setNavOrder` method.
        """

        actions.ActionProvider     .__init__(self, overlayList, displayCtx)
        props.SyncableHasProperties.__init__(self)

        if not isinstance(displayCtx, displaycontext.DisplayContext):
            raise TypeError(
                'displayCtx must be a '
                '{} instance'.format( displaycontext.DisplayContext.__name__))

        self.__frame     = frame
        self.__name      = '{}_{}'.format(self.__class__.__name__, id(self))
        self.__destroyed = False
        self.__navOrder  = None

        if kbFocus:
            self.Bind(wx.EVT_CHAR_HOOK, self.__onCharHook)


    def setNavOrder(self, children):
        """Set the keyboard (tab, shift+tab) navigation order to the
        given list of controls, assumed to be children of this
        ``_FSLeyesPanel``.
        """

        nav = []

        allChildren = []
        for c in children:
            if type(c) == wx.Panel: allChildren.extend(c.GetChildren())
            else:                   allChildren.append(c)

        children = allChildren

        log.debug('Updating nav order for {}'.format(type(self).__name__))
        for i, w in enumerate(children):

            log.debug('{} nav {:2d}: {}'.format(
                type(self).__name__,
                i,
                type(w).__name__))

            # Special cases for some of our custom controls
            if isinstance(w, floatspin.FloatSpinCtrl):
                nav.append(w.textCtrl)
            elif isinstance(w, floatslider.SliderSpinPanel):
                nav.append(w.spinCtrl.textCtrl)
            elif isinstance(w, rangeslider.RangePanel):

                low  = w.lowWidget
                high = w.highWidget

                if isinstance(low, floatspin.FloatSpinCtrl):
                    low = low.textCtrl
                if isinstance(high, floatspin.FloatSpinCtrl):
                    high = high.textCtrl

                nav.extend([low, high])

            elif isinstance(w, rangeslider.RangeSliderSpinPanel):
                nav.extend([w.lowSpin.textCtrl, w.highSpin.textCtrl])
            else:
                nav.append(w)

        self.__navOrder = nav


    def __onCharHook(self, ev):
        """Called on ``EVT_CHAR_HOOK`` events. Intercepts tab key presses,
        to force an explicit keyboard navigation ordering.
        """

        if self.__navOrder is None or ev.GetKeyCode() != wx.WXK_TAB:
            ev.Skip()
            return

        # Get the widget that has focus
        try:
            focusIdx = self.__navOrder.index(wx.Window.FindFocus())

            log.debug('{} focus nav event ({:2d} [{}] is focused)'.format(
                type(self).__name__,
                focusIdx,
                type(wx.Window.FindFocus()).__name__))

        # Some other widget that we
        # don't care about has focus.
        except Exception:
            ev.Skip()
            return

        if ev.ShiftDown(): offset = -1
        else:              offset =  1

        # Get the next widget in
        # the tab traversal order
        nextIdx = (focusIdx + offset) % len(self.__navOrder)

        # Search for the next enabled widget
        while not (self.__navOrder[nextIdx].IsEnabled() and
                   self.__navOrder[nextIdx].IsShownOnScreen()):

            if nextIdx == focusIdx:
                break

            nextIdx = (nextIdx + offset) % len(self.__navOrder)

        toFocus = self.__navOrder[nextIdx]

        log.debug('{}: moving focus to {:2d} [{}]'.format(
            type(self).__name__,
            nextIdx,
            type(toFocus).__name__))

        toFocus.SetFocus()

        # If the next widget to receive
        # focus is a TextCtrl, select
        # all of its text
        if isinstance(toFocus, wx.TextCtrl):
            toFocus.SelectAll()


    @property
    def name(self):
        """Returns a unique name associated with this ``_FSLeyesPanel``. """
        return self.__name


    @property
    def frame(self):
        """Returns the :class:`.FSLeyesFrame` which created this
        ``_FSLeyesPanel``. May be ``None``, if this panel was not created
        by a ``FSLeyesFrame``.
        """
        return self.__frame


    def destroy(self):
        """This method must be called by whatever is managing this
        ``_FSLeyesPanel`` when it is to be closed/destroyed.

        It seems to be impossible to define a single handler (on either the
        :attr:`wx.EVT_CLOSE` and/or :attr:`wx.EVT_WINDOW_DESTROY` events)
        which handles both cases where the window is destroyed (in the process
        of destroying a parent window), and where the window is explicitly
        closed by the user (e.g. when embedded as a page in a Notebook).

        This issue is probably caused by my use of the AUI framework for
        layout management, as the AUI manager/notebook classes do not seem to
        call close/destroy in all cases. Everything that I've tried, which
        relies upon ``EVT_CLOSE``/``EVT_WINDOW_DESTROY`` events, inevitably
        results in the event handlers not being called, or in segmentation
        faults (presumably due to double-frees at the C++ level).

        Subclasses which need to perform any cleaning up when they are closed
        may override this method, and should be able to assume that it will be
        called. So this method *must* be called by managing code when a panel
        is deleted.

        Overriding subclass implementations must call this base class method,
        otherwise memory leaks will probably occur, and warnings will probably
        be output to the log (see :meth:`__del__`). This implememtation should
        be called **after** the subclass has performed its own clean-up, as
        this method expliciltly clears the ``overlayList`` and ``displayCtx``
        references (via :meth:`.ActionProvider.destroy`).
        """
        actions.ActionProvider.destroy(self)
        self.__frame       = None
        self.__navOrder    = None
        self.__destroyed   = True


    @property
    def destroyed(self):
        """Returns ``True`` if a call to :meth:`destroy` has been made,
        ``False`` otherwise.
        """
        return self.__destroyed


    def __del__(self):
        """If the :meth:`destroy` method has not been called, a warning message
        is logged.
        """
        if not self.__destroyed:
            log.warning('The {}.destroy() method has not been called '
                        '- unless the application is shutting down, '
                        'this is probably a bug!'.format(type(self).__name__))


class FSLeyesPanel(_FSLeyesPanel, wx.Panel):
    """The ``FSLeyesPanel`` is the base class for all view and control panels
    in *FSLeyes*. See the :mod:`fsleyes` documentation for more details.

    See also the :class:`.ViewPanel` and :class:`.ControlPanel` classes.
    """

    __metaclass__ = sip.wrappertype


    def __init__(self,
                 parent,
                 overlayList,
                 displayCtx,
                 frame,
                 *args,
                 **kwargs):

        # Slightly ugly way of supporting the _FSLeyesPanel
        # kbFocus argument. In order to catch keyboard events,
        # we need the WANTS_CHARS style. So we peek in
        # kwargs to see if it is present. If it is, we add
        # WANTS_CHARS to the style flag.
        kbFocus = kwargs.pop('kbFocus', False)
        if kbFocus:

            # The wx.Panel style defaults to TAB_TRAVERSAL
            style = kwargs.get('style', wx.TAB_TRAVERSAL)
            kwargs['style'] = style | wx.WANTS_CHARS

        wx.Panel.__init__(self, parent, *args, **kwargs)
        _FSLeyesPanel.__init__(self, overlayList, displayCtx, frame, kbFocus)



#!/usr/bin/env python
#
# applyflirtxfm.py - The ApplyFlirtXfmAction class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`ApplyFlirtXfmAction` class, an
:class:`.Action` which allows the user to load a FLIRT transformation
matrix and apply it to an :class:`.Image` overlay.


A number of related standalone classes and functions are also defined in this
module:

   .. autosummary::
      :nosignatures:

      FlirtFileDialog
      promptForFlirtFiles
      guessFlirtFiles
      calculateTransform
"""


import            os
import os.path as op

import numpy as np

import wx

import fsl.data.image       as fslimage
import fsl.transform.flirt  as flirt
import fsleyes.strings      as strings
import fsleyes.actions.base as base


class ApplyFlirtXfmAction(base.NeedOverlayAction):
    """The ``ApplyFlirtXfmAction`` class is an action which allows the user to
    load a FLIRT transformation matrix (or other affine file) and apply it to
    the currently selected overlay, if it is an :class:`.Image` instance.


    A :class:`FlirtFileDialog` is used to prompt the user to select a
    transformation matrix and reference image. The :func:`calculateTransform`
    function is used to calculate the source voxel -> reference world
    transformation, and the image ``voxToWorldMat`` is then updated
    accordingly.
    """


    def __init__(self, overlayList, displayCtx, frame):
        """Create an ``ApplyFlirtXfmAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg frame:       The :class:`.FSLeyesFrame`.
        """
        base.NeedOverlayAction.__init__(
            self, overlayList, displayCtx, func=self.__applyFlirtXfm)
        self.__frame = frame


    def __applyFlirtXfm(self):
        """Called when this action is executed.
        """

        displayCtx  = self.displayCtx
        overlayList = self.overlayList
        overlay     = displayCtx.getSelectedOverlay()

        affType, matFile, refFile = promptForFlirtFiles(
            self.__frame, overlay, overlayList, displayCtx)

        if all((affType is None, matFile is None, refFile is None)):
            return

        if affType == 'flirt':
            xform = calculateTransform(
                overlay, overlayList, displayCtx, matFile, refFile)
        elif affType == 'v2w':
            xform = np.loadtxt(matFile)

        overlay.voxToWorldMat = xform


def calculateTransform(overlay, overlayList, displayCtx, matFile, refFile):
    """Calculates a source voxel -> reference world transformation matrix
    from the given FLIRT transform and refernece image files.

    See the :func:`fsl.transform.flirt.flirtMatrixToSform` function.

    :arg overlay:     The :class:`.Image` overlay - the source image of the
                      FLIRT transformation.

    :arg overlayList: The :class:`.OverlayList`.

    :arg displayCtx:  The :class:`.DisplayContext`.

    :arg matFile:     Path to the FLIRT transformation matrix file.

    :arg refFile:     Path to the FLIRT reference image file.
    """

    # The reference image might
    # already be in the overlay list.
    refImg = overlayList.find(refFile)

    if refImg is None:
        refImg = fslimage.Image(refFile)

    flirtMat = np.loadtxt(matFile)

    return flirt.flirtMatrixToSform(flirtMat, overlay, refImg)


def promptForFlirtFiles(parent, overlay, overlayList, displayCtx, save=False):
    """Displays a dialog prompting the user to select a FLIRT
    transformation matrix file and associated reference image for
    the given overlay.

    :arg parent:      The :mod:`wx` parent object.
    :arg overlay:     The overlay to load a FLIRT matrix for.
    :arg overlayList: The :class:`.OverlayList` instance.
    :arg displayCtx:  The :class:`.DisplayContext` instance.
    :arg save:        Prompt the user to save a transformation matrix instead.

    :returns: A tuple containing:

              - The affine type  currently one of ``'flirt'``, indicating a
                FLIRT matrix, or ``'v2w'``, indicating a "raw" voxel-to-world
                matrix.
              - The selected matrix file.
              - The selected reference image file (``None`` if
                ``affType is 'v2w'``)

             If the user cancelled the dialog, all elements of this tuple will
             be ``None``.
    """

    if overlay.dataSource is not None:
        matFile, refFile = guessFlirtFiles(overlay.dataSource)
    else:
        matFile, refFile = None, None

    refOptFiles = []
    refOpts     = []
    selectedRef = None

    for ovl in reversed(overlayList):

        if not isinstance(ovl, fslimage.Image): continue

        refOptFiles.append(ovl.dataSource)
        refOpts    .append(displayCtx.getDisplay(ovl).name)

        # If a reference file has been
        # guesed, make sure it is
        # selected in the drop down box
        if refFile is not None and ovl.dataSource == refFile:
            selectedRef = len(refOpts) - 1

    if len(refOpts) == 0:
        refOpts     = None
        refOptFiles = None

    dlg = FlirtFileDialog(parent,
                          overlay.dataSource,
                          refOpts=refOpts,
                          refOptFiles=refOptFiles,
                          selectedRef=selectedRef,
                          matFile=matFile,
                          refFile=refFile,
                          save=save)

    dlg.Layout()
    dlg.Fit()
    dlg.CentreOnParent()

    if dlg.ShowModal() == wx.ID_OK:
        return dlg.GetAffineType(), dlg.GetMatFile(), dlg.GetRefFile()
    else:
        return None, None, None


def guessFlirtFiles(path):
    """Given a ``path`` to a NIFTI image file, tries to guess an appropriate
    FLIRT transformation matrix file and reference image. The guess is based
    on the path location (e.g. if it is a FEAT or MELODIC image).

    Returns a tuple containing paths to the matrix file and reference image,
    or ``(None, None)`` if a guess couldn't be made.
    """

    import fsl.data.featanalysis    as featanalysis
    import fsl.data.melodicanalysis as melodicanalysis

    if path is None:
        return None, None

    filename  = op.basename(path)
    dirname   = op.dirname( path)
    regDir    = None
    srcRefMap = {}

    func2struc = 'example_func2highres.mat'
    struc2std  = 'highres2standard.mat'

    featDir = featanalysis   .getAnalysisDir(dirname)
    melDir  = melodicanalysis.getAnalysisDir(dirname)

    # TODO more heuristics
    if featDir is not None:

        regDir    = op.join(featDir, 'reg')
        srcRefMap = {
            'example_func'            : ('highres',  func2struc),
            'filtered_func_data'      : ('highres',  func2struc),
            'mean_func'               : ('highres',  func2struc),
            'thresh_zstat'            : ('highres',  func2struc),
            op.join('stats', 'zstat') : ('highres',  func2struc),
            op.join('stats', 'tstat') : ('highres',  func2struc),
            op.join('stats', 'cope')  : ('highres',  func2struc),
            op.join('stats', 'pe')    : ('highres',  func2struc),
            'highres'                 : ('standard', struc2std),
            'highres_head'            : ('standard', struc2std),
            'struc'                   : ('standard', struc2std),
            'struct'                  : ('standard', struc2std),
            'struct_brain'            : ('standard', struc2std),
            'struc_brain'             : ('standard', struc2std),
        }

    elif melodicanalysis.isMelodicDir(dirname):

        if melDir.startswith('filtered_func_data'):
            regDir = op.join(melDir, '..', 'reg')
        else:
            regDir = op.join(melDir, 'reg')

        srcRefMap = {
            'filtered_func_data' : ('highres',  func2struc),
            'melodic_IC'         : ('highres',  func2struc),
            'example_func'       : ('highres',  func2struc),
            'mean_func'          : ('highres',  func2struc),
            'highres'            : ('standard', struc2std),
            'highres_head'       : ('standard', struc2std),
            'struc'              : ('standard', struc2std),
            'struct'             : ('standard', struc2std),
            'struct_brain'       : ('standard', struc2std),
            'struc_brain'        : ('standard', struc2std),
        }

    matFile = None
    refFile = None

    for src, (ref, mat) in srcRefMap.items():

        if not filename.startswith(src):
            continue

        mat = op.join(regDir, mat)
        ref = op.join(regDir, ref)

        try:              ref = fslimage.addExt(ref)
        except Exception: continue

        if op.exists(mat):
            matFile = mat
            refFile = ref
            break

    return matFile, refFile


class FlirtFileDialog(wx.Dialog):
    """The ``FlirtFileDialog`` class is a ``wx.Dialog`` which prompts the user to
    select a FLIRT (or other affine) transformation matrix and reference image
    associated with a source image.

    The user can select a reference image either from a drop down box, or
    by selecting a file in the file system.
    """


    def __init__(self,
                 parent,
                 srcFile,
                 refOpts=None,
                 refOptFiles=None,
                 selectedRef=None,
                 matFile=None,
                 refFile=None,
                 save=False):
        """Create a ``FlirtFileDialog``.

        :arg parent:      The ``wx`` parent object.

        :arg srcFile:     Path to the FLIRT source image file

        :arg refOpts:     Options to use in the reference image drop down box.

        :arg refOptFiles: File paths which correspond to the ``refOpts``.

        :arg selectedRef: Index of initially selected ``refOpt``.

        :arg matFile:     Initial path to a FLIRT transformation matrix file

        :arg refFile:     Initial Path to a FLIRT reference image file

        :arg save:        If ``True``, the user will be prompted to save a
                          FLIRT matrix. Otherwise (the default), the user will
                          be prompted to load an existing FLIRT matrix.
        """

        titleKey = {True : 'save', False : 'load'}[save]

        wx.Dialog.__init__(self,
                           parent,
                           title=strings.titles[self, titleKey],
                           style=(wx.DEFAULT_DIALOG_STYLE | wx.STAY_ON_TOP))

        if refOpts     is None: refOpts     = []
        if refOptFiles is None: refOptFiles = []
        if selectedRef is None: selectedRef = 0

        self.__srcFile     = srcFile
        self.__refOpts     = list(refOpts)
        self.__refOptFiles = list(refOptFiles)
        self.__matFile     = None
        self.__refFile     = None
        self.__save        = save
        self.__affineTypes = ['flirt', 'v2w']

        affTypeOpts = [strings.labels[self, 'affType', at]
                       for at in self.__affineTypes]

        refOpts = list(refOpts) + [strings.labels[self, 'refChoiceSelectFile']]

        overlayName    = wx.StaticText(self, style=wx.ALIGN_CENTRE_HORIZONTAL)
        label          = wx.StaticText(self, style=wx.ALIGN_CENTRE_HORIZONTAL)
        affType        = wx.Choice(self, choices=affTypeOpts)
        affTypeLabel   = wx.StaticText(self)
        refChoiceLabel = wx.StaticText(self)
        matFileLabel   = wx.StaticText(self)
        refFileLabel   = wx.StaticText(self)
        refChoice      = wx.Choice(self, choices=refOpts)
        matFileText    = wx.TextCtrl(self)
        refFileText    = wx.TextCtrl(self)
        matFileButton  = wx.Button(self)
        refFileButton  = wx.Button(self)
        okButton       = wx.Button(self, wx.ID_OK)
        cancelButton   = wx.Button(self, wx.ID_CANCEL)

        self.__matFileText    = matFileText
        self.__refFileText    = refFileText
        self.__refFileLabel   = refFileLabel
        self.__refFileButton  = refFileButton
        self.__refChoiceLabel = refChoiceLabel
        self.__refChoice      = refChoice
        self.__affType        = affType

        if srcFile is None:
            srcName = strings.labels[self, 'inmemory']
        else:
            srcName = op.basename(srcFile)
            srcName = fslimage.removeExt(srcFile)

        overlayName   .SetLabel(strings.labels[self, 'source'].format(srcName))
        affTypeLabel  .SetLabel(strings.labels[self, 'affType'])
        refChoiceLabel.SetLabel(strings.labels[self, 'refImage'])
        matFileLabel  .SetLabel(strings.labels[self, 'matFile'])
        refFileLabel  .SetLabel(strings.labels[self, 'refFile'])
        matFileButton .SetLabel(strings.labels[self, 'selectFile'])
        refFileButton .SetLabel(strings.labels[self, 'selectFile'])
        okButton      .SetLabel(strings.labels[self, 'ok'])
        cancelButton  .SetLabel(strings.labels[self, 'cancel'])
        refChoice     .SetSelection(selectedRef)
        affType       .SetSelection(0)

        if save: label.SetLabel(strings.labels[self, 'save.message'])
        else:    label.SetLabel(strings.labels[self, 'load.message'])

        if matFile is not None:
            matFileText.SetValue(matFile)
            matFileText.SetInsertionPointEnd()
        if refFile is not None:
            refFileText.SetValue(refFile)
            refFileText.SetInsertionPointEnd()

        sizer       = wx.BoxSizer(wx.VERTICAL)
        widgetSizer = wx.FlexGridSizer(4, 3, 0, 0)
        btnSizer    = wx.BoxSizer(wx.HORIZONTAL)

        widgetSizer.AddGrowableCol(1)

        sizer.Add((1, 10),     flag=wx.EXPAND)
        sizer.Add(overlayName, flag=wx.ALIGN_CENTRE)
        sizer.Add((1, 10),     flag=wx.EXPAND)
        sizer.Add(label,       flag=wx.ALIGN_CENTRE)
        sizer.Add((1, 10),     flag=wx.EXPAND)
        sizer.Add(widgetSizer, flag=wx.EXPAND | wx.LEFT | wx.RIGHT, border=10)
        sizer.Add((1, 10),     flag=wx.EXPAND)
        sizer.Add(btnSizer,    flag=wx.EXPAND | wx.LEFT | wx.RIGHT, border=10)
        sizer.Add((1, 10),     flag=wx.EXPAND)

        widgetSizer.Add(affTypeLabel, flag=wx.ALL,    border=3)
        widgetSizer.Add(affType,      flag=wx.EXPAND, proportion=1)
        widgetSizer.Add((-1, -1))

        if refOpts is not None:
            widgetSizer.Add(refChoiceLabel, flag=wx.ALL,    border=3)
            widgetSizer.Add(refChoice,      flag=wx.EXPAND, proportion=1)
            widgetSizer.Add((-1, -1))
        else:
            widgetSizer.Add((-1, -1))
            widgetSizer.Add((-1, -1))
            widgetSizer.Add((-1, -1))

        widgetSizer.Add(refFileLabel,  flag=wx.ALL,    border=3)
        widgetSizer.Add(refFileText,   flag=wx.EXPAND, proportion=1)
        widgetSizer.Add(refFileButton, flag=wx.EXPAND)
        widgetSizer.Add(matFileLabel,  flag=wx.ALL,    border=3)
        widgetSizer.Add(matFileText,   flag=wx.EXPAND, proportion=1)
        widgetSizer.Add(matFileButton, flag=wx.EXPAND)

        btnSizer.Add((10, 1),      flag=wx.EXPAND, proportion=1)
        btnSizer.Add(okButton,     flag=wx.EXPAND)
        btnSizer.Add((10, 1),      flag=wx.EXPAND)
        btnSizer.Add(cancelButton, flag=wx.EXPAND)
        btnSizer.Add((10, 1),      flag=wx.EXPAND, proportion=1)

        matFileText.SetMinSize((400, -1))
        refFileText.SetMinSize((400, -1))

        okButton.SetDefault()
        self.SetSizer(sizer)

        okButton     .Bind(wx.EVT_BUTTON, self.__onOkButton)
        cancelButton .Bind(wx.EVT_BUTTON, self.__onCancelButton)
        matFileButton.Bind(wx.EVT_BUTTON, self.__onMatFileButton)
        refFileButton.Bind(wx.EVT_BUTTON, self.__onRefFileButton)
        affType      .Bind(wx.EVT_CHOICE, self.__onAffType)
        refChoice    .Bind(wx.EVT_CHOICE, self.__onRefChoice)

        self.__okButton     = okButton
        self.__cancelButton = cancelButton
        self.__matFileText  = matFileText
        self.__refFileText  = refFileText
        self.__affType      = affType
        self.__refChoice    = refChoice

        if len(refOpts) == 1:
            refChoice     .Disable()
            refChoiceLabel.Disable()
        else:
            self.__onRefChoice(None)


    @property
    def ok(self):
        """Return a reference to the OK button. """
        return self.__okButton


    @property
    def cancel(self):
        """Return a reference to the cancel button. """
        return self.__cancelButton


    @property
    def matFileText(self):
        """Return a reference to the matrix file text entry widget. """
        return self.__matFileText


    @property
    def refFileText(self):
        """Return a reference to the reference file text entry widget. """
        return self.__refFileText


    @property
    def affType(self):
        """Return a reference to the affine type dropdown widget. """
        return self.__affType


    @property
    def refChoice(self):
        """Return a reference to the reference file dropdown widget. """
        return self.__refChoice


    def GetAffineType(self):
        """Return the currently selected affine type - currently either
        ``'flirt'`` (indicating a FSL FLIRT matrix file), or ``'v2w'``
        (indicating a "raw" voxel-to-world affine).
        """
        return self.__affineTypes[self.__affType.GetSelection()]


    def GetMatFile(self):
        """Returns the current value of the matrix file as a string, or
        ``None``, if the file path is not valid.
        """

        matFile = self.__matFileText.GetValue()

        if self.__save or op.exists(matFile): return op.abspath(matFile)
        else:                                 return None


    def GetRefFile(self):
        """Returns the current value of the reference file, as a string, or
        ``None``, if the file path is not valid.
        """

        choice = self.__refChoice.GetSelection()

        if choice < len(self.__refOpts):
            refFile = self.__refOptFiles[choice]
        else:
            refFile = self.__refFileText.GetValue()

        if op.exists(refFile): return op.abspath(refFile)
        else:                  return None


    def __onOkButton(self, ev):
        """Called when the user clicks the ok button. Closes the dialog.
        """
        self.EndModal(wx.ID_OK)


    def __onCancelButton(self, ev):
        """Called when the user clicks the cancel button. Closes the dialog.
        """
        self.__matFileText.SetValue('')
        self.__refFileText.SetValue('')
        self.EndModal(wx.ID_CANCEL)


    def __onAffType(self, ev):
        """Called when the user changes the affine type. Enables/disables
        widgets related to the reference image (as they are only used for
        FLIRT affines).
        """

        affType = self.__affineTypes[self.__affType.GetSelection()]
        isFlirt = affType == 'flirt'

        self.__refChoice     .Enable(isFlirt)
        self.__refChoiceLabel.Enable(isFlirt)
        self.__refFileLabel  .Enable(isFlirt)
        self.__refFileText   .Enable(isFlirt)
        self.__refFileButton .Enable(isFlirt)


    def __onRefChoice(self, ev):
        """Called when the user changes the selection in the reference image
        drop down box. Enables/disables the reference image file selection
        widgets as necessary.
        """

        choice     = self.__refChoice.GetSelection()
        selectFile = choice >= len(self.__refOpts)

        self.__refFileLabel .Enable(selectFile)
        self.__refFileText  .Enable(selectFile)
        self.__refFileButton.Enable(selectFile)


    def __onMatFileButton(self, ev):
        """Called when the user clicks the matrix file select button.
        Displays a file dialog prompting the user to select a matrix file.
        """

        if self.__save: style = wx.FD_SAVE
        else:           style = wx.FD_OPEN

        matFile = self.__matFileText.GetValue().strip()
        if matFile == '':
            if self.__srcFile is None: dirName = os.getcwd()
            else:                      dirName = op.dirname(self.__srcFile)
            fileName = 'xform.mat'
        else:
            dirName, fileName = op.split(matFile)

        dlg = wx.FileDialog(
            self,
            defaultDir=dirName,
            defaultFile=fileName,
            message=strings.messages[self, 'matFile'],
            style=style)

        if dlg.ShowModal() == wx.ID_OK:
            self.__matFileText.SetValue(dlg.GetPath())


    def __onRefFileButton(self, ev):
        """Called when the user clicks the reference file select button.
        Displays a file dialog prompting the user to select a reference file.
        """

        refFile = self.__refFileText.GetValue()
        if refFile == '':
            if self.__srcFile is None: refFile = ''
            else:                      refFile = self.__srcFile

        dlg = wx.FileDialog(
            self,
            defaultFile=refFile,
            message=strings.messages[self, 'refFile'],
            style=wx.FD_OPEN)

        if dlg.ShowModal() == wx.ID_OK:
            self.__refFileText.SetValue(dlg.GetPath())



#!/usr/bin/env python
#
# __init__.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""The ``displaycontext`` package contains classes which define the display
options for pretty much everything in *FSLeyes*.


.. note:: Before perusing this package, you should read the high level
          overview in the :mod:`~fsl.fsleyes` package documentation. Go on -
          it won't take you too long.


--------
Overview
--------


The most important classes defined in this package are:

  - the :class:`.DisplayContext` class, which defines how all of the
    overlays in an :class:`.OverlayList` should be displayed.

  - the :class:`.Display` class, which defines how a single overlay should be
    displayed.

  - the :class:`.DisplayOpts` base class, and its sub-classes, which define
    overlay type specific options.


A :class:`.DisplayContext` instance encapsulates an :class:`.OverlayList`, and
defines how the overlays in the list should be displayed.  Each
:class:`.ViewPanel` displayed in *FSLeyes* (e.g. the :class:`.OrthoPanel`) has
its own ``DisplayContext`` instance; a ``ViewPanel`` uses its
``DisplayContext`` instance to configure general display properties, and also
to access the :class:`.Display` properties for individual overlays.


All of the classes mentioned on ths page are defined in sub-modules, but are
imported into, and are thus available from, the ``displaycontext`` package
namespace. For example::

    import fsleyes.displaycontext as fsldc

    # The VolumeOpts class is defined in the
    # fsleyes.displaycontext.volumeopts
    # module, but is available in the
    # fsleyes.displaycontext namespace.
    volopts = fsldc.VolumeOpts(overlay, display, overlayList, displayCtx)


-----------------------
Overlay display options
-----------------------


The :class:`.Display` class, and the :class:`.DisplayOpts` sub-classes define
how to display a single overlay. Options common to all overlays
(e.g. :attr:`.Display.brightness`, :attr:`.Display.alpha`) are defined in the
``Display`` class, whereas options which are specific to a particular overlay
type (e.g. :attr:`.VolumeOpts.cmap`, :attr:`.LineVectorOpts.lineWidth`) are
defined in the corresponding :class:`.DisplayOpts` sub-class.


The ``Display`` instance for a particular overlay owns and manages a single
``DisplayOpts`` instance - whenever the overlay display type is changed, the
``Display`` instance deletes the old ``DisplayOpts`` instance, and creates a
new one accordingly.  The following ``DisplayOpts`` sub-classes exist:

.. autosummary::
   :nosignatures:

   ~fsleyes.displaycontext.niftiopts.NiftiOpts
   ~fsleyes.displaycontext.volumeopts.VolumeOpts
   ~fsleyes.displaycontext.volumeopts.VolumeRGBOpts
   ~fsleyes.displaycontext.volumeopts.ComplexOpts
   ~fsleyes.displaycontext.volume3dopts.Volume3DOpts
   ~fsleyes.displaycontext.maskopts.MaskOpts
   ~fsleyes.displaycontext.vectoropts.RGBVectorOpts
   ~fsleyes.displaycontext.vectoropts.LineVectorOpts
   ~fsleyes.displaycontext.meshopts.MeshOpts
   ~fsleyes.displaycontext.giftiopts.GiftiOpts
   ~fsleyes.displaycontext.freesurferopts.FreesurferOpts
   ~fsleyes.displaycontext.labelopts.LabelOpts
   ~fsleyes.displaycontext.tensoropts.TensorOpts
   ~fsleyes.displaycontext.shopts.SHOpts
   ~fsleyes.displaycontext.mipopts.MIPOpts
   ~fsleyes.displaycontext.tractogramopts.TractogramOpts


--------------
Overlay groups
--------------


.. note:: Support for overlay groups is quite basic at this point in time. See
          the :class:`.OverlayListPanel` for details.

The :mod:`~.displaycontext.group` module provides the functionality to
link the display properties of one or more overlays. One or more
:class:`.OverlayGroup` instances may be added to the
:attr:`.DisplayContext.overlayGroups` list.


-------------
Scene options
-------------


Independent of the ``DisplayContext``, ``Display`` and ``DisplayOpts``
classes, the ``displaycontext`` package is also home to a few classes which
define *scene* options:


.. autosummary::
   :nosignatures:

   ~fsleyes.displaycontext.canvasopts.SliceCanvasOpts
   ~fsleyes.displaycontext.canvasopts.LightBoxCanvasOpts
   ~fsleyes.displaycontext.canvasopts.Scene3DCanvasOpts
   ~fsleyes.displaycontext.sceneopts.SceneOpts
   ~fsleyes.displaycontext.orthoopts.OrthoOpts
   ~fsleyes.displaycontext.lightboxopts.LightBoxOpts
   ~fsleyes.displaycontext.scene3dopts.Scene3DOpts


.. note:: Aside from an increase in code modularity and cleanliness, another
          reason that all of these scene display settings are separated from
          the things that use them is so they can be imported, queried, and
          set without having to import the modules that use them.

          For example, the :mod:`.parseargs` module needs to be able
          to access all of the display settings in order to print out the
          corresponding help documentation. This process would take much
          longer if ``parseargs`` had to import, e.g.
          :class:`.OrthoPanel`, which would result in importing both :mod:`wx`
          and :mod:`.OpenGL`.

          Keeping these display settings separate allows us to avoid these
          time-consuming imports in situations where they are not needed.
"""


import itertools as it

import fsleyes_widgets.utils.typedict as td
import fsl.data.constants             as constants

from .               import display
from .displaycontext import DisplayContext
from .display        import Display
from .display        import DisplayOpts
from .group          import OverlayGroup
from .sceneopts      import SceneOpts
from .orthoopts      import OrthoOpts
from .lightboxopts   import LightBoxOpts
from .scene3dopts    import Scene3DOpts
from .colourmapopts  import ColourMapOpts
from .niftiopts      import NiftiOpts
from .volumeopts     import VolumeOpts
from .volumeopts     import ComplexOpts
from .volumeopts     import VolumeRGBOpts
from .volume3dopts   import Volume3DOpts
from .maskopts       import MaskOpts
from .vectoropts     import VectorOpts
from .vectoropts     import NiftiVectorOpts
from .vectoropts     import RGBVectorOpts
from .vectoropts     import LineVectorOpts
from .meshopts       import MeshOpts
from .giftiopts      import GiftiOpts
from .freesurferopts import FreesurferOpts
from .labelopts      import LabelOpts
from .tensoropts     import TensorOpts
from .shopts         import SHOpts
from .mipopts        import MIPOpts
from .tractogramopts import TractogramOpts

from .displaycontext import InvalidOverlayError


OVERLAY_TYPES = td.TypeDict({

    'Image'          : ['volume',     'mask',  'rgbvector',
                        'linevector', 'label', 'sh',
                        'tensor',     'mip',   'rgb',
                        'complex'],
    'Mesh'           : ['mesh'],
    'DTIFitTensor'   : ['tensor', 'rgbvector', 'linevector'],
    'Tractogram'     : ['tractogram']
})
"""This dictionary provides a mapping between all overlay classes,
and the possible values that the :attr:`Display.overlayType` property
may take for each of them.

For each overlay class, the first entry in the corresponding overlay type
list is used as the default overlay type.
"""


ALL_OVERLAY_TYPES = list(set(it.chain(*OVERLAY_TYPES.values())))
"""This attribute contains a list of all possible overlay types - see the
:attr:`.Display.overlayType` property and tge :data:`.display.OVERLAY_TYPES`
dictionary for more details.
"""


DISPLAY_OPTS_MAP = td.TypeDict({
    'Nifti.volume'          : VolumeOpts,
    'Nifti.rgbvector'       : RGBVectorOpts,
    'Nifti.linevector'      : LineVectorOpts,
    'Nifti.mask'            : MaskOpts,
    'Nifti.label'           : LabelOpts,
    'Nifti.tensor'          : TensorOpts,
    'Nifti.sh'              : SHOpts,
    'Nifti.mip'             : MIPOpts,
    'Nifti.rgb'             : VolumeRGBOpts,
    'Nifti.complex'         : ComplexOpts,
    'Mesh.mesh'             : MeshOpts,
    'VTKMesh.mesh'          : MeshOpts,
    'GiftiMesh.mesh'        : GiftiOpts,
    'FreesurferMesh.mesh'   : FreesurferOpts,
    'Tractogram.tractogram' : TractogramOpts,
})
"""This dictionary provides a mapping between each (overlay type,
:attr:`.Display.overlayType`) pair, and the :class:`DisplayOpts` subclass
which contains overlay type-specific display options.
"""


def getOverlayTypes(overlay):
    """Returns a list of possible overlay types for the given overlay.
    This is a wrapper around the :attr:`OVERLAY_TYPES` dictionary, which
    might adjust the returned list based on properties of the overlay.
    """

    import fsl.data.image as fslimage
    from . import            shopts

    possibleTypes = list(OVERLAY_TYPES[overlay])

    if not isinstance(overlay, fslimage.Image):
        return possibleTypes

    shape     = overlay.shape
    ndims     = len(shape)
    nvals     = overlay.nvals
    iscomplex = overlay.iscomplex

    # Could this image be a vector image?
    couldBeVector = ((ndims == 4 and shape[-1] == 3) or
                     (ndims == 3 and nvals     == 3))

    # Could this image be a RGB(A) image?
    couldBeRGB = (nvals in (3, 4)) or (ndims == 4 and shape[-1] in (3, 4))

    # Or could it be a SH or tensor image?
    couldBeTensor = ndims == 4 and shape[-1] == 6
    couldBeSH     = ndims == 4 and shape[-1] in shopts.SH_COEFFICIENT_TYPE

    # Special cases

    # If the image is complex, make complex the
    # default overlay type
    possibleTypes.remove('complex')
    if iscomplex:
        possibleTypes.insert(0, 'complex')

    # If the overlay looks like a vector image,
    # and its nifti intent code is set as such,
    # make rgbvector the default overlay type
    if couldBeVector:
        if overlay.intent == constants.NIFTI_INTENT_RGB_VECTOR:
            possibleTypes.remove(   'rgbvector')
            possibleTypes.remove(   'linevector')
            possibleTypes.insert(0, 'linevector')
            possibleTypes.insert(0, 'rgbvector')

    # Otherwise, remove the vector options
    else:
        try:               possibleTypes.remove('rgbvector')
        except ValueError: pass

        try:               possibleTypes.remove('linevector')
        except ValueError: pass

    if not couldBeRGB:
        try:               possibleTypes.remove('rgb')
        except ValueError: pass

    if not couldBeSH:
        try:               possibleTypes.remove('sh')
        except ValueError: pass

    if not couldBeTensor:
        try:               possibleTypes.remove('tensor')
        except ValueError: pass

    return possibleTypes



#!/usr/bin/env python
#
# about.py - The AboutDialog class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`.AboutDialog` class, a dialog which
displays information about *FSLeyes*.
"""


import os.path as op

import wx

from   fsl.utils.platform import platform as fslplatform
import fsleyes_widgets.imagepanel         as imagepanel
import fsleyes.gl                         as fslgl
import fsleyes.strings                    as strings
import fsleyes.splash                     as splash
import fsleyes.version                    as version


class AboutDialog(wx.Dialog):
    """The ``AboutDialog`` is a dialog which displays information about
    *FSLeyes*.
    """

    def __init__(self, parent):
        """Create an ``AboutDialog``.

        :arg parent: ``wx`` parent object.
        """
        wx.Dialog.__init__(self, parent, title=strings.about['title'])

        # Load the splash screen
        splashbmp = wx.Bitmap(splash.getSplashFile(), wx.BITMAP_TYPE_PNG)
        splashimg = splashbmp.ConvertToImage()

        # Create all the widgets
        splashPanel = imagepanel.ImagePanel(self, splashimg)
        splashPanel.SetMinSize(splashimg.GetSize())

        textPanel   = wx.TextCtrl(self,
                                  size=(-1, 200),
                                  style=(wx.TE_LEFT      |
                                         wx.TE_RICH      |
                                         wx.TE_MULTILINE |
                                         wx.TE_READONLY  |
                                         wx.TE_AUTO_URL))
        closeButton = wx.Button(self, id=wx.ID_CANCEL)

        # Set foreground/background colours
        textPanel.SetBackgroundColour('#000000')
        textPanel.SetForegroundColour('#ffffff')
        textPanel.SetDefaultStyle(wx.TextAttr('#ffffff', '#000000'))

        # Create / retrieve all the content
        verStr    = version.__version__
        glVerStr  = fslgl.GL_VERSION
        glRenStr  = fslgl.GL_RENDERER
        glCompat  = fslgl.GL_COMPATIBILITY
        swlibs    = strings.about['libs']

        swVersions = []
        for lib in swlibs:

            try:
                mod = __import__(lib)

                # Retrieve the sub-module if necessary
                for n in lib.split(".")[1:]:
                    mod = getattr(mod, n)

                if lib == 'PIL':
                    try:
                        swVer = str(mod.__version__)
                    except Exception:
                        swVer = str(mod.PILLOW_VERSION)
                else:
                    swVer = str(mod.__version__)
            except Exception:
                swVer = ''

            swVersions.append(swVer)

        verStr    = strings.about['version']   .format(verStr)
        fslVer    = strings.about['fslVersion'].format(fslplatform.fslVersion)
        fslDir    = strings.about['fslPath']   .format(fslplatform.fsldir)
        glVerStr  = strings.about['glVersion'] .format(glVerStr)
        glCompat  = strings.about['glCompat']  .format(glCompat)
        glRenStr  = strings.about['glRenderer'].format(glRenStr)
        swStr     = strings.about['software']  .format(*swVersions)

        # Tack the license file contents onto
        # the end of the software description.
        licenseFile = op.join(op.dirname(__file__),
                              '..', '..', 'LICENSE')
        try:
            with open(licenseFile, 'rt') as f:
                licenseStr = f.read()
        except Exception:
            licenseStr = ''

        swStr = swStr + '\n\n' + licenseStr
        swStr = swStr.strip()

        # Set the widget content
        infoStr = '\n'.join((verStr,
                             strings.about['company'],
                             strings.about['author'],
                             strings.about['email'],
                             fslVer,
                             fslDir,
                             glVerStr,
                             glCompat,
                             glRenStr))

        textPanel  .SetValue(infoStr + '\n\n' + swStr)
        closeButton.SetLabel('Close')

        # Arrange the widgets
        sizer = wx.BoxSizer(wx.VERTICAL)

        sizer.Add(splashPanel)
        sizer.Add(textPanel, flag=wx.EXPAND, proportion=1)
        sizer.Add(closeButton, flag=wx.EXPAND)

        self.SetSizer(sizer)
        self.Layout()
        self.Fit()



#!/usr/bin/env python
#
# globject.py - The GLObject class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`GLObject` class, which is a superclass
for all FSLeyes OpenGL overlay types. The following classes are
defined in this module:

.. autosummary::
   :nosignatures:

   GLObject
   GLSimpleObject


See also the :class:`.GLImageObject`, which is the base class for all
``GLObject`` sub-types that display :class:`.Nifti` overlays.


This module also provides a few functions, most importantly
:func:`createGLObject`:

.. autosummary::
   :nosignatures:

   getGLObjectType
   createGLObject
"""


import contextlib
import logging

import fsl.utils.idle     as idle
import fsl.utils.notifier as notifier


log = logging.getLogger(__name__)


def getGLObjectType(overlayType):
    """This function returns an appropriate :class:`GLObject` type for the
    given :attr:`.Display.overlayType` value.
    """

    from . import glvolume
    from . import glrgbvolume
    from . import glcomplex
    from . import glmask
    from . import glrgbvector
    from . import gllinevector
    from . import glmesh
    from . import gllabel
    from . import gltensor
    from . import glsh
    from . import glmip
    from . import gltractogram

    typeMap = {
        'volume'     : glvolume    .GLVolume,
        'mask'       : glmask      .GLMask,
        'rgbvector'  : glrgbvector .GLRGBVector,
        'linevector' : gllinevector.GLLineVector,
        'mesh'       : glmesh      .GLMesh,
        'label'      : gllabel     .GLLabel,
        'tensor'     : gltensor    .GLTensor,
        'sh'         : glsh        .GLSH,
        'mip'        : glmip       .GLMIP,
        'rgb'        : glrgbvolume .GLRGBVolume,
        'complex'    : glcomplex   .GLComplex,
        'tractogram' : gltractogram.GLTractogram
    }

    return typeMap.get(overlayType, None)


def createGLObject(overlay, overlayList, displayCtx, threedee=False):
    """Create :class:`GLObject` instance for the given overlay, as specified
    by the :attr:`.Display.overlayType` property.

    :arg overlay:     An overlay object (e.g. a :class:`.Image` instance).

    :arg overlayList: The :class:`.OverlayList`

    :arg displayCtx:  The :class:`.DisplayContext` managing the scene.

    :arg threedee:    If ``True``, the ``GLObject`` will be configured for
                      3D rendering. Otherwise it will be configured for 2D
                      slice-based rendering.
    """

    display = displayCtx.getDisplay(overlay)
    ctr     = getGLObjectType(display.overlayType)

    if ctr is not None:
        return ctr(overlay, overlayList, displayCtx, threedee)
    else:
        return None


class GLObject(notifier.Notifier):
    """The :class:`GLObject` class is a base class for all OpenGL objects
    displayed in *FSLeyes*.


    **Instance attributes**


    The following attributes will always be available on ``GLObject``
    instances:

      - ``name``:       A unique name for this ``GLObject`` instance.

      - ``overlay``:    The overlay to be displayed.

      - ``display``:    The :class:`.Display` instance describing the
                        overlay display properties.

      - ``opts``:       The :class:`.DisplayOpts` instance describing the
                        overlay-type specific display properties.

      - ``displayCtx``: The :class:`.DisplayContext` managing the scene
                        that this ``GLObject`` is a part of.

      - ``threedee``:   A boolean flag indicating whether this ``GLObject``
                        is configured for 2D or 3D rendering.


    **Usage**


    Once you have created a ``GLObject``:

     1. Do not use the ``GLObject`` until its :meth:`ready` method returns
        ``True``.

     2. In order to render the ``GLObject`` to a canvas, call (in order) the
        :meth:`preDraw`, :meth:`draw2D` (or :meth:`draw3D`), and
        :meth:`postDraw`, methods. Multple calls to
        :meth:`draw2D`/:meth:`draw3D` may occur between calls to
        :meth:`preDraw` and :meth:`postDraw`.

     3. Once you are finished with the ``GLObject``, call its :meth:`destroy`
        method.


    Note that a ``GLObject`` which has been created for 2D rendering
    is not expected be able to render in 3D, nor vice-versa.


    **Update listeners**


    A ``GLObject`` instance will notify registered listeners when its state
    changes and it needs to be re-drawn.  Entities which are interested in
    changes to a ``GLObject`` instance may register as *update listeners*, via
    the :meth:`.Notifier.register` method. It is the resposibility of
    sub-classes of ``GLObject`` to call the :meth:`.Notifier.notify` method to
    facilitate this notification process.


    **Sub-class resposibilities***


    Sub-class implementations must do the following:

     - Call :meth:`__init__`. A ``GLObject.__init__`` sub-class method must
       have the following signature, and must pass all arguments through to
       ``GLObject.__init__``::

           def __init__(self, overlay, displayCtx, canvas, threedee)

     - Call :meth:`notify` whenever its OpenGL representation changes.

     - Override the following methods:

       .. autosummary::
          :nosignatures:

          getDisplayBounds
          getDataResolution
          ready
          destroy
          destroyed
          preDraw
          draw2D
          draw3D
          postDraw

    Alternately, a sub-class could derive from one of the following classes,
    instead of deriving directly from the ``GLObject`` class:

    .. autosummary::
       :nosignatures:

       GLSimpleObject
       .GLImageObject
    """


    def __init__(self, overlay, overlayList, displayCtx, threedee):
        """Create a :class:`GLObject`.  The constructor adds one attribute
        to this instance, ``name``, which is simply a unique name for this
        instance.

        Subclass implementations must call this method, and should also
        perform any necessary OpenGL initialisation, such as creating
        textures.

        :arg overlay:     The overlay

        :arg overlayList: The :class:`.OverlayList`

        :arg displayCtx:  The ``DisplayContext`` managing the scene

        :arg threedee:    Whether this ``GLObject`` is to be used for 2D or 3D
                          rendering.
        """

        self.__name        = '{}_{}'.format(type(self).__name__, id(self))
        self.__threedee    = threedee
        self.__overlay     = overlay
        self.__overlayList = overlayList
        self.__displayCtx  = displayCtx
        self.__display     = None
        self.__opts        = None

        # GLSimpleObject passes in None for
        # both the overlay and the displayCtx.
        if overlay is not None and displayCtx is not None:
            self.__display = displayCtx.getDisplay(overlay)
            self.__opts    = self.__display.opts

        log.debug('{}.init ({})'.format(type(self).__name__, id(self)))


    def __del__(self):
        """Prints a log message."""
        if log:
            log.debug('{}.del ({})'.format(type(self).__name__, id(self)))


    @property
    def name(self):
        """A unique name for this ``GLObject``. """
        return self.__name


    @property
    def overlay(self):
        """The overlay being drawn by this ``GLObject``."""
        return self.__overlay


    @property
    def display(self):
        """The :class:`.Display` instance containing overlay display
        properties.
        """
        return self.__display


    @property
    def opts(self):
        """The :class:`.DisplayOpts` instance containing overlay
        (type-specific) display properties.
        """
        return self.__opts


    @property
    def overlayList(self):
        """The :class:`.OverlayList`."""
        return self.__overlayList


    @property
    def displayCtx(self):
        """The :class:`.DisplayContext` dsecribing thef scene that this
        ``GLObject`` is a part of.
        """
        return self.__displayCtx


    @property
    def threedee(self):
        """Property which is ``True`` if this ``GLObject`` was configured
        for 3D rendering, or ``False`` if it was configured for 2D slice
        rendering.
        """
        return self.__threedee


    def ready(self):
        """This method must return ``True`` or ``False`` to indicate
        whether this ``GLObject`` is ready to be drawn. The method should,
        for example, make sure that all :class:`.ImageTexture` objects
        are ready to be used.
        """
        raise NotImplementedError('The ready method must be '
                                  'implemented by GLObject subclasses')


    def notifyWhen(self, condition):
        """Wrapper around :meth:`.Notifier.notify` which schedule the
        ``notify`` call to take place when ``condition() is True``.
        """
        name = f'{self.name}_notifyWhen'
        idle.idleWhen(self.notify, condition, name=name, skipIfQueued=True)


    def getDisplayBounds(self):
        """This method must calculate and return a bounding box, in the
        display coordinate system, which contains the entire ``GLObject``.
        The bounds must be returned as a tuple with the following structure::

            ((xlo, ylo, zlo), (xhi, yhi, zhi))

        This method must be implemented by sub-classes.
        """

        raise NotImplementedError('The getDisplayBounds method must be '
                                  'implemented by GLObject subclasses')


    def getBoundsLengths(self):
        """Convenience method which returns a tuple containing the
        ``(x, y, z)``  lengths of the bounding box which contains the
        ``GLObject``.
        """
        los, his = self.getDisplayBounds()
        return tuple([hi - lo for lo, hi in zip(los, his)])


    def getDataResolution(self, xax, yax):
        """This method must calculate and return a sequence of three values,
        which defines a suitable pixel resolution, along the display coordinate
        system ``(x, y, z)`` axes, for rendering a 2D slice of this
        ``GLObject`` to screen.


        This method should be implemented by sub-classes. If not implemented,
        a default resolution is used. The returned resolution *might* be used
        to render this ``GLObject``, but typically only in a low performance
        environment where off-screen rendering to a
        :class:`.GLObjectRenderTexture` is used - see the
        :class:`.SliceCanvas` documentation for more details.


        :arg xax: Axis to be used as the horizontal screen axis.
        :arg yax: Axis to be used as the vertical screen axis.
        """
        return None


    def destroy(self):
        """This method must be called when this :class:`GLObject` is no longer
        needed.

        It should perform any necessary cleaning up, such as deleting texture
        objects.

        .. note:: Sub-classes which override this method must call this
                  implementation.
        """
        self.__overlay    = None
        self.__display    = None
        self.__opts       = None
        self.__displayCtx = None


    @property
    def destroyed(self):
        """This method may be called to test whether a call has been made to
        :meth:`destroy`.

        It should return ``True`` if this ``GLObject`` has been destroyed,
        ``False`` otherwise.
        """
        raise NotImplementedError()


    def preDraw(self):
        """This method is called at the start of a draw routine.

        It should perform any initialisation which is required before one or
        more calls to the :meth:`draw2D`/:meth:`draw3D` methods are made, such
        as binding and configuring textures.
        """


    def draw2D(self, canvas, zpos, axes, xform=None):
        """This method is called on ``GLObject`` instances which are
        configured for 2D rendering. It should draw a view of this
        ``GLObject`` - a 2D slice at the given Z location, which specifies
        the position along the screen depth axis.

        :arg canvas: The canvas which is displaying this ``GLObject``.
                     Could be a :class:`.SliceCanvas`, a
                     :class:`.LightBoxCanvas`, a :class:`.Scene3DCanvas`,
                     or some future not-yet-created canvas.

        :arg zpos:   Position along Z axis to draw.

        :arg axes:   Tuple containing the ``(x, y, z)`` axes in the
                     display coordinate system The ``x`` and ``y`` axes
                     correspond to the horizontal and vertical display axes
                     respectively, and the ``z`` to the depth.

        :arg xform:  If provided, it must be applied to the model view
                     transformation before drawing.
        """


    def draw3D(self, canvas, xform=None):
        """This method is called on ``GLObject`` instances which are
        configured for 3D rendering. It should draw a 3D view of this
        ``GLObject``.

        :arg canvas: The canvas which is displaying this ``GLObject``.
                     Could be a :class:`.SliceCanvas`, a
                     :class:`.LightBoxCanvas`, a :class:`.Scene3DCanvas`,
                     or some future not-yet-created canvas.

        :arg xform:  If provided, it must be applied to the model view
                     transformation before drawing.
        """


    def drawAll(self, canvas, axes, zposes, xforms):
        """This is a convenience method for 2D lightboxD canvases, where
        multple 2D slices at different depths are drawn alongside each other.

        This method should do the same as multiple calls to the :meth:`draw2D`
        method, one for each of the Z positions and transformation matrices
        contained in the ``zposes`` and ``xforms`` arrays (``axes`` is fixed).

        In some circumstances (hint: the :class:`.LightBoxCanvas`), better
        performance may be achieved in combining multiple renders, rather
        than doing it with separate calls to :meth:`draw`.

        The default implementation does exactly this, so this method need only
        be overridden for subclasses which are able to get better performance
        by combining the draws.
        """
        for (zpos, xform) in zip(zposes, xforms):
            self.draw2D(canvas, zpos, axes, xform)


    def postDraw(self):
        """This method is called after the :meth:`draw2D`/:meth:`draw3D`
        methods have been called one or more times.

        It should perform any necessary cleaning up, such as unbinding
        textures.
        """


class GLSimpleObject(GLObject):
    """The ``GLSimpleObject`` class is a convenience superclass for simple
    rendering tasks (probably fixed-function) which are not associated with a
    specific overlay, and require no setup or initialisation/management of GL
    memory or state. It is used by the :mod:`.annotations` module.

    All subclasses need to do is implement the :meth:`GLObject.draw2D` and
    :meth:`GLObject.draw3D` methods. The :mod:`.annotations` module uses the
    ``GLSimpleObject`` class.

    Subclasses should not assume that any of the other methods will ever
    be called.

    .. note:: The :attr:`GLObject.overlay`, :attr:`GLObject.display`,
              :attr:`GLObject.opts`, :attr:`GLObject.canvas`,
              :attr:`GLObject.overlayList` and :attr:`GLObject.displayCtx`
              properties of a ``GLSimpleObject`` will all potentially be
              set to ``None``.
    """


    def __init__(self, threedee):
        """Create a ``GLSimpleObject``. """
        GLObject.__init__(self, None, None, None, threedee)
        self.__destroyed = False


    def ready(self):
        """Overrides :meth:`GLObject.ready`. Returns ``True``. """
        return True


    def destroy( self):
        """Overrides :meth:`GLObject.destroy`. Does nothing. """
        GLObject.destroy(self)
        self.__destroyed = True


    @property
    def destroyed(self):
        """Overrides :meth:`GLObject.destroy`. Returns ``True`` if
        :meth:`destroy` hs been called, ``False`` otherwise.
        """
        return self.__destroyed


    def preDraw(self):
        """Overrides :meth:`GLObject.preDraw`. Does nothing. """


    def postDraw(self):
        """Overrides :meth:`GLObject.postDraw`. Does nothing. """



#!/usr/bin/env python
#
# glvector.py - The GLVectorBase and GLVector classes.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`GLVectorBase` and :class:`GLVector`
classes. The ``GLVectorBase`` class encapsulate the logic for rendering
overlays which contain directional data, and the ``GLVector`` class
specifically conatins logic for displaying :class:`.Image` overlays with
shape ``X*Y*Z*3``, or of type ``NIFTI_TYPE_RGB24``.
"""


import functools            as ft

import numpy                as np
import OpenGL.GL            as gl

import fsl.utils.idle       as idle
import fsleyes.colourmaps   as fslcm
from . import resources     as glresources
from . import                  textures
from . import                  glimageobject


class GLVectorBase(glimageobject.GLImageObject):
    """The :class:`GLVectorBase` class encapsulates the logic for rendering
    :class:`.Nifti` overlay types which represent directional data (and which
    are described by a :class:`.VectorOpts` instance).  The ``GLVectorBase``
    class is a sub-class of :class:`.GLImageObject`.


    The ``GLVectorBase`` class is a base class which is not intended to be
    instantiated directly. The :class:`.GLRGBVector`, :class:`.GLLineVector`,
    :class:`.GLTensor`, and :class:`.GLSH` subclasses should be used instead.
    These subclasses share the functionality provided by this class. See also
    the :class:`GLVector` class, which is also a base class.


    *Colouring*

    A ``GLVectorBase`` can be coloured in one of two ways:

     - Each voxel is coloured according to the orientation of the vector.
       A custom fragment shader program looks up the ``xyz`` vector values,
       and combines three colours (corresponding to the ``xyz`` directions)
       to form the final fragment colour. The colours for each component
       are specified by the :attr:`.VectorOpts.xColour`,
       :attr:`.VectorOpts.yColour`, and :attr:`.VectorOpts.zColour`
       properties. If the image being displayed contains directional data
       (e.g. is a ``X*Y*Z*3`` vector image), you should use the
       :class:`GLVector` class.

     - Each voxel is coloured according to the values contained in another
       image, which are used to look up a colour in a colour map. The image
       and colour map are respectively specified by the
       :attr:`.VectorOpts.colourImage` and :attr:`.VectorOpts.cmap` properties.


    In either case, the brightness or transparency of each vector colour may
    be modulated by another image, specified by the
    :attr:`.VectorOpts.modulateImage` and :attr:`.VectorOpts.modulateMode`
    properties.  This modulation image is stored as a 3D single-channel
    :class:`.ImageTexture`.


    Finally, vector voxels may be clipped according to the values of another
    image, specified by the :attr:`.VectorOpts.clipImage` property.  This
    clipping image is stored as a 3D single-channel :class:`.ImageTexture`, and
    the clipping thresholds specified by the :attr:`.VectorOpts.clippingRange`
    property.


    *Textures*

    The ``GLVectorBase`` class configures its textures in the following manner:

    =================== ==================
    ``modulateTexture`` ``gl.GL_TEXTURE0``
    ``clipTexture``     ``gl.GL_TEXTURE1``
    ``colourTexture``   ``gl.GL_TEXTURE2``
    ``cmapTexture``     ``gl.GL_TEXTURE3``
    =================== ==================
    """


    def __init__(self,
                 overlay,
                 overlayList,
                 displayCtx,
                 threedee,
                 init=None,
                 preinit=None):
        """Create a ``GLVectorBase`` object bound to the given overlay and
        display.

        Initialises the OpenGL data required to render the given vector
        overlay. This method does the following:

          - Creates the modulate, clipping and colour image textures.

          - Adds listeners to the :class:`.Display` and :class:`.VectorOpts`
            instances, so the textures and geometry can be updated when
            necessary.

        :arg overlay:     A :class:`.Nifti` object.

        :arg overlayList: The :class:`.OverlayList`

        :arg displayCtx:  A :class:`.DisplayContext` object which describes
                          how the overlay is to be displayed.

        :arg threedee:    2D or 3D rendering.

        :arg init:        An optional function to be called when all of the
                          :class:`.ImageTexture` instances associated with
                          this ``GLVectorBase`` have been initialised.

        :arg preinit:     An optional functiono be called after this
                          ``GLVectorBase`` has configured itself, but
                          *before* ``init`` is called. Used by
                          :class:`GLVector`.
        """

        glimageobject.GLImageObject.__init__(self,
                                             overlay,
                                             overlayList,
                                             displayCtx,
                                             threedee)

        name = self.name
        opts = self.opts

        self.cmapTexture = textures.ColourMapTexture('{}_cm'.format(name))
        self.shader      = None
        self.auxmgr      = textures.AuxImageTextureManager(
            self, colour=None, modulate=None, clip=None)

        self.registerAuxImage('colour',   opts.colourImage)
        self.registerAuxImage('modulate', opts.modulateImage)
        self.registerAuxImage('clip',     opts.clipImage)

        self.addListeners()
        self.refreshColourMapTexture()

        def initWrapper():
            # Sometimes (e.g. during testing),
            # this GLObject might get destroyed
            # before this function is called
            if self.destroyed:
                return
            if init is not None:
                init()
            self.notify()

        if preinit is not None:
            preinit()

        idle.idleWhen(initWrapper, self.texturesReady)


    def destroy(self):
        """Must be called when this ``GLVectorBase`` is no longer needed.
        Deletes the GL textures, and deregisters the listeners configured in
        :meth:`__init__`.
        """

        self.cmapTexture.destroy()
        self.auxmgr.destroy()
        self.removeListeners()

        self.cmapTexture = None
        self.auxmgr      = None

        glimageobject.GLImageObject.destroy(self)


    @property
    def modulateTexture(self):
        """Returns the :class:`.ImageTexture` for the
        :attr:`.VectorOpts.modulateImage`.
        """
        return self.auxmgr.texture('modulate')


    @property
    def clipTexture(self):
        """Returns the :class:`.ImageTexture` for the
        :attr:`.VectorOpts.clipImage`.
        """
        return self.auxmgr.texture('clip')


    @property
    def colourTexture(self):
        """Returns the :class:`.ImageTexture` for the
        :attr:`.VectorOpts.colourImage`.
        """
        return self.auxmgr.texture('colour')


    def ready(self):
        """Returns ``True`` if this ``GLVectorBase`` is ready to be drawn,
        ``False`` otherwise.
        """
        return self.shader is not None and self.texturesReady()


    def texturesReady(self):
        """Calls :meth:`.AuxImageTextureManager.texturesReady`. """
        return self.auxmgr.texturesReady()


    def addListeners(self):
        """Called by :meth:`__init__`. Adds listeners to properties of the
        :class:`.Display` and :class:`.VectorOpts` instances, so that the GL
        representation can be updated when the display properties change.
        """

        display = self.display
        opts    = self.opts
        name    = self.name

        display.addListener('alpha',         name, self.__cmapPropChanged)
        display.addListener('brightness',    name, self.__cmapPropChanged)
        display.addListener('contrast',      name, self.__cmapPropChanged)
        opts   .addListener('xColour',       name, self.asyncUpdateShaderState)
        opts   .addListener('yColour',       name, self.asyncUpdateShaderState)
        opts   .addListener('zColour',       name, self.asyncUpdateShaderState)
        opts   .addListener('suppressX',     name, self.asyncUpdateShaderState)
        opts   .addListener('suppressY',     name, self.asyncUpdateShaderState)
        opts   .addListener('suppressZ',     name, self.asyncUpdateShaderState)
        opts   .addListener('suppressMode',  name, self.asyncUpdateShaderState)
        opts   .addListener('modulateMode',  name, self.asyncUpdateShaderState)
        opts   .addListener('cmap',          name, self.__cmapPropChanged)
        opts   .addListener('modulateImage', name, self.__modImageChanged)
        opts   .addListener('clipImage',     name, self.__clipImageChanged)
        opts   .addListener('colourImage',   name, self.__colourImageChanged)
        opts   .addListener('clippingRange', name, self.asyncUpdateShaderState)
        opts   .addListener('modulateRange', name, self.asyncUpdateShaderState)
        opts   .addListener('transform',     name, self.notify)


    def removeListeners(self):
        """Called by :meth:`destroy`. Removes all property listeners added
        by  the :meth:`addListeners` method.
        """

        display = self.display
        opts    = self.opts
        name    = self.name

        display.removeListener('alpha',         name)
        display.removeListener('brightness',    name)
        display.removeListener('contrast',      name)
        opts   .removeListener('xColour',       name)
        opts   .removeListener('yColour',       name)
        opts   .removeListener('zColour',       name)
        opts   .removeListener('cmap',          name)
        opts   .removeListener('suppressX',     name)
        opts   .removeListener('suppressY',     name)
        opts   .removeListener('suppressZ',     name)
        opts   .removeListener('suppressMode',  name)
        opts   .removeListener('modulateMode',  name)
        opts   .removeListener('modulateImage', name)
        opts   .removeListener('clipImage',     name)
        opts   .removeListener('colourImage',   name)
        opts   .removeListener('clippingRange', name)
        opts   .removeListener('modulateRange', name)
        opts   .removeListener('transform' ,    name)


    def compileShaders(self):
        """This method must be provided by subclasses (e.g.g the
        :class:`.GLRGBVector` and :class:`.GLLineVector` classes), and must
        compile the vertex/fragment shaders used to render this
        ``GLVectorBase``.
        ."""
        raise NotImplementedError('compileShaders must be implemented by '
                                  '{} subclasses'.format(type(self).__name__))


    def updateShaderState(self):
        """This method must be provided by subclasses (e.g. the
        :class:`.GLRGBVector` and :class:`.GLLineVector` classes), and must
        update the state of the vertex/fragment shader programs. It must return
        ``True`` if the shader state was updated, ``False`` otherwise.
        """
        raise NotImplementedError('updateShaderState must be implemented by '
                                  '{} subclasses'.format(type(self).__name__))


    def asyncUpdateShaderState(self, *args, **kwargs):
        """Calls :meth:`updateShaderState` and then :meth:`.Notifier.notify`, using
        :func:`.idle.idleWhen` function to make sure that it is only called
        when :meth:`ready` returns ``True``.
        """

        alwaysNotify = kwargs.pop('alwaysNotify', None)

        def func():
            if self.updateShaderState() or alwaysNotify:
                self.notify()

        idle.idleWhen(func,
                      self.ready,
                      name=self.name,
                      skipIfQueued=True)


    def refreshColourMapTexture(self, colourRes=256):
        """Called when the component colour maps need to be updated, when one
        of the :attr:`.VectorOpts.xColour`, ``yColour``, ``zColour``, ``cmap``,
        ``suppressX``, ``suppressY``, or ``suppressZ`` properties change.

        Regenerates the colour map texture.
        """

        display = self.display
        opts    = self.opts

        if opts.colourImage is not None:
            dmin, dmax = opts.colourImage.dataRange

        else:
            dmin, dmax = 0.0, 1.0

        dmin, dmax = fslcm.briconToDisplayRange(
            (dmin, dmax),
            display.brightness / 100.0,
            display.contrast   / 100.0)

        self.cmapTexture.set(cmap=opts.cmap,
                             alpha=display.alpha / 100.0,
                             displayRange=(dmin, dmax))


    def getClippingRange(self):
        """Returns the :attr:`clippingRange`, suitable for use in the fragment
        shader. The returned values are transformed into the clip image
        texture value range, so the fragment shader can compare texture
        values directly to it.
        """

        opts = self.opts

        clipLow, clipHigh = opts.clippingRange
        xform             = self.clipTexture.invVoxValXform

        if opts.clipImage is not None:
            clipLow  = clipLow  * xform[0, 0] + xform[0, 3]
            clipHigh = clipHigh * xform[0, 0] + xform[0, 3]
        else:
            clipLow  = -0.1
            clipHigh =  1.1

        return clipLow, clipHigh


    def getModulateRange(self):
        """Returns the :attr:`modulateRange`, suitable for use in the fragment
        shader. The returned values are transformed into the modulate image
        texture value range, so the fragment shader can compare texture values
        directly to it.
        """

        opts = self.opts

        modLow, modHigh = opts.modulateRange
        xform           = self.modulateTexture.invVoxValXform

        if opts.modulateImage is not None:
            modLow  = modLow  * xform[0, 0] + xform[0, 3]
            modHigh = modHigh * xform[0, 0] + xform[0, 3]
        else:
            modLow  = 0
            modHigh = 1

        return modLow, modHigh


    def getAuxTextureXform(self, which):
        """Generates and returns a transformation matrix which can be used
        to transform texture coordinates from the vector image to the specified
        auxillary image (``'clip'``, ``'modulate'`` or ``'colour'``).
        """
        return self.auxmgr.textureXform(which)


    def preDraw(self):
        """Must be called by subclass implementations.

        Ensures that all of the textures managed by this ``GLVectorBase`` are
        bound to their corresponding texture units.
        """

        self.modulateTexture.bindTexture(gl.GL_TEXTURE0)
        self.clipTexture    .bindTexture(gl.GL_TEXTURE1)
        self.colourTexture  .bindTexture(gl.GL_TEXTURE2)
        self.cmapTexture    .bindTexture(gl.GL_TEXTURE3)


    def postDraw(self):
        """Must be called by subclass implementations.

        Unbinds all of the textures managed by this ``GLVectorBase``.
        """

        self.modulateTexture.unbindTexture()
        self.clipTexture    .unbindTexture()
        self.colourTexture  .unbindTexture()
        self.cmapTexture    .unbindTexture()


    def __cmapPropChanged(self, *a):
        """Called when a :class:`.Display` or :class:`.VectorOpts` property
        affecting the vector colour map settings changes. Calls
        :meth:`refreshColourMapTexture` and :meth:`asyncUpdateShaderState`.
        """
        self.refreshColourMapTexture()
        self.asyncUpdateShaderState(alwaysNotify=True)


    def registerAuxImage(self, which, image, onReady=None, **kwargs):
        """Registers the given auxillary image with the
        ``AuxImageTextureManager``.
        """

        self.auxmgr.texture(which).deregister(self.name)
        self.auxmgr.registerAuxImage(which, image, **kwargs)
        self.auxmgr.texture(which).register(self.name, self.__textureChanged)
        if onReady is not None:
            idle.idleWhen(onReady, self.auxmgr.texturesReady)


    def __colourImageChanged(self, *a):
        """Called when the :attr:`.VectorOpts.colourImage` changes. Registers
        with the new image, and refreshes textures as needed.
        """
        def onReady():
            self.compileShaders()
            self.refreshColourMapTexture()
            self.asyncUpdateShaderState(alwaysNotify=True)
        self.registerAuxImage('colour', self.opts.colourImage, onReady)


    def __modImageChanged(self, *a):
        """Called when the :attr:`.VectorOpts.modulateImage` changes.
        Registers with the new image, and refreshes textures as needed.
        """
        onReady = ft.partial(self.asyncUpdateShaderState, alwaysNotify=True)
        self.registerAuxImage('modulate', self.opts.modulateImage, onReady)


    def __clipImageChanged(self, *a):
        """Called when the :attr:`.VectorOpts.clipImage` changes.
        Registers with the new image, and refreshes textures as needed.
        """
        onReady = ft.partial(self.asyncUpdateShaderState, alwaysNotify=True)
        self.registerAuxImage('clip', self.opts.clipImage, onReady)


    def __textureChanged(self, *a):
        """Called when any of the :class:`.ImageTexture` instances containing
        clipping, modulation or colour data, are refreshed. Notifies
        listeners of this ``GLVectorBase`` (via the :class:`.Notifier` base
        class).
        """
        self.asyncUpdateShaderState(alwaysNotify=True)


class GLVector(GLVectorBase):
    """The ``GLVector`` class is a sub-class of :class:`GLVectorBase`, which
    contains some additional logic for rendering :class:`.Image` overlays with
    a shape ``X*Y*Z*3``, or of type ``NIFTI_TYPE_RGB24``, and which contain
    directional data.


    By default , the ``image`` overlay passed to :meth:`__init__` is assumed
    to be an :class:`.Image` instance which contains vector data. If this is
    not the case, the ``vectorImage`` parameter may be used to pass in the
    :class:`.Image` that contains the vector data.


    This vector image is stored on the GPU as a 3D RGB :class:`.ImageTexture`,
    where the ``R`` channel contains the ``x`` vector values, the ``G``
    channel the ``y`` values, and the ``B`` channel the ``z`` values.


    This texture is bound to texture unit  ``gl.GL_TEXTURE4`` in the
    :meth:`preDraw` method.
    """

    def __init__(self, image, *args, **kwargs):
        """Create a ``GLVector``. All of the arguments documented here are
        optional, but if provided, must be passed as keyword arguments. All
        other arguments are passed through to :meth:`GLVectorBase.__init__`.

        The ``image``, (or ``vectorImage``) argument is assumed to be an
        :class:`.Image` instance of shape ``(X, Y, Z, 3)``, or of type
        ``NIFTI_TYPE_RGB24``, which contains the vector data. If the former,
        the vector data is assumed to be in the range ``[-1, 1]``. If the
        latter, the vector data is, by definition, in the range ``[0, 255]``
        - this is assumed to map directly to the range ``[-1, 1]``.


        :arg vectorImage:    If ``None``, the ``image`` is assumed to be an
                             :class:`.Image` instance which contains the
                             vector data. If this is not the case, the
                             ``vectorImage`` parameter can be used to specify
                             an ``Image`` instance which does contain the
                             vector data.

        :arg prefilter:      An optional function which filters the data before
                             it is stored as a 3D texture. See
                             :class:`.Texture3D`. Regardless of whether this
                             function is provided, the data is always
                             transposed so that the fourth dimension is the
                             fastest changing, before being transferred to the
                             GPU.

        :arg prefilterRange: If the provided ``prefilter`` function will cause
                             the range of the data to change, this function
                             must be provided, and must, given the original
                             data range, return a suitably adjusted adjust data
                             range.
        """

        def defaultPrefilter(d):
            return d

        vectorImage    = kwargs.pop('vectorImage',    image)
        prefilter      = kwargs.pop('prefilter',      defaultPrefilter)
        prefilterRange = kwargs.pop('prefilterRange', None)

        # Must be an image of shape (X, Y, Z, 3),
        # or of type RGB24. If the latter, the
        # test below will accept any numpy
        # structured array with three values per
        # voxel, but we assume elsewhere that
        # those values are all of type np.uint8
        # (and thus correspond to the
        # NIFTI_TYPE_RGB24 type)
        shape = vectorImage.shape
        ndims = len(shape)
        nvals = vectorImage.nvals
        isRGB = (((ndims == 4) and (nvals == 1) and (shape[3] == 3)) or
                 ((ndims == 3) and (nvals == 3)))

        if not isRGB:
            raise ValueError('Image must be 4 dimensional with 3 volumes '
                             'representing the XYZ vector angles, or of '
                             'type RGB24')

        self.vectorImage     = vectorImage
        self.imageTexture    = None
        self.prefilter       = prefilter
        self.prefilterRange  = prefilterRange

        # Using the preinit hook to overcome a slight
        # chicken-and-egg problem. We need to create
        # the image texture before shaders are created
        # (which are done via the init hook, as defined
        # in sub-classes, e.g. GLRGBVector). But we
        # need access to the display/displayopts objects
        # in order to configure the image texture.
        # So pre-init ensures that the GLObject refs
        # (display/displayOpts) are set up, then the
        # image texture is refreshed, then the init hook
        # is called.
        preinit = kwargs.pop('preinit', None)

        def preinitWrapper():
            self.refreshImageTexture()
            if preinit is not None:
                preinit()

        GLVectorBase.__init__(self,
                              image,
                              *args,
                              preinit=preinitWrapper,
                              **kwargs)


    def destroy(self):
        """Overrides :meth:`GLVectorBase.destroy`. Must be called when this
        ``GLVector`` is no longer needed. Calls :meth:`GLVectorBase.destroy`,
        and destroys the vector image texture.
        """

        GLVectorBase.destroy(self)
        self.imageTexture.deregister(self.name)
        glresources.delete(self.imageTexture.name)

        self.imageTexture = None


    def texturesReady(self):
        """Overrides :meth:`GLVectorBase.texturesReady`.  Returns ``True`` if
        all of the textures managed by this ``GLVector`` are ready to be used,
        ``False`` otherwise.
        """
        return (self.imageTexture is not None and
                self.imageTexture.ready()     and
                GLVectorBase.texturesReady(self))


    def refreshImageTexture(self, interp=gl.GL_NEAREST):
        """Called by :meth:`__init__`, and when the :class:`.ImageTexture`
        needs to be updated. (Re-)creates the ``ImageTexture``, using the
        :mod:`.resources` module so that the texture can be shared by other
        users.

        :arg interp: Interpolation method (``GL_NEAREST`` or ``GL_LINEAR``).
                     Used by sub-class implementations (see
                     :class:`.GLRGBVector`).
        """

        prefilter      = self.prefilter
        prefilterRange = self.prefilterRange
        vecImage       = self.vectorImage
        texName        = '{}_{}'.format(type(self).__name__, id(vecImage))

        if self.imageTexture is not None:
            self.imageTexture.deregister(self.name)
            glresources.delete(self.imageTexture.name)

        self.imageTexture = glresources.get(
            texName,
            textures.ImageTexture,
            texName,
            vecImage,
            nvals=3,
            interp=interp,
            normaliseRange=vecImage.dataRange,
            prefilter=prefilter,
            prefilterRange=prefilterRange,
            notify=False)

        self.imageTexture.register(self.name, self.__textureChanged)


    def preDraw(self):
        """Overrides :meth:`GLVectorBase`. Binds the vector image texture.
        """
        GLVectorBase.preDraw(self)
        self.imageTexture.bindTexture(gl.GL_TEXTURE4)


    def postDraw(self):
        """Overrides :meth:`GLVectorBase`. Unbinds the vector image texture.
        """
        GLVectorBase.postDraw(self)
        self.imageTexture.unbindTexture()


    def __textureChanged(self, *a):
        """Called when the :class:`.ImageTexture` instance containing the vector
        data is are refreshed. Notifies listeners of this ``GLVector`` (via the
        :class:`.Notifier` base class).
        """
        self.asyncUpdateShaderState(alwaysNotify=True)



#!/usr/bin/env python
#
# componentgrid.py - the ComponentGrid class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`ComponentGrid` class, which is used by
the :class:`.MelodicClassificationPanel`.
"""


import logging

import wx

import fsl.data.image             as fslimage
import fsl.utils.idle             as idle

import fsleyes_props              as props
import fsleyes_widgets.widgetgrid as widgetgrid
import fsleyes_widgets.texttag    as texttag
import fsleyes.panel              as fslpanel
import fsleyes.strings            as strings
import fsleyes.displaycontext     as fsldisplay


log = logging.getLogger(__name__)


class ComponentGrid(fslpanel.FSLeyesPanel):
    """The ``ComponentGrid`` uses a :class:`.WidgetGrid`, and a set of
    :class:`.TextTagPanel` widgets, to display the component classifications
    stored in the :class:`.VolumeLabels` object that is associated
    with an :class:`.Image` (typically a :class:`.MelodicImage`). The
    ``Image`` and ``VolumeLabels`` instance is specified via the
    :meth:`setOverlay` method.


    The grid contains one row for each component, and a ``TextTagPanel`` is
    used to display the labels associated with each component. Each
    ``TextTagPanel`` allows the user to add and remove labels to/from the
    corresponding component.


    See also the :class:`.LabelGrid` class, which displays the same
    information, but organised by label.
    """


    def __init__(self, parent, overlayList, displayCtx, frame, lut):
        """Create a ``ComponentGrid``.

        :arg parent:      The ``wx`` parent object.
        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg frame:       The :class:`.FSLeyesFrame` instance.
        :arg lut:         The :class:`.LookupTable` instance used to colour
                          each label tag.
        """

        fslpanel.FSLeyesPanel.__init__(
            self, parent, overlayList, displayCtx, frame)

        self.__lut  = lut
        self.__grid = widgetgrid.WidgetGrid(
            self,
            style=(wx.VSCROLL                    |
                   widgetgrid.WG_SELECTABLE_ROWS |
                   widgetgrid.WG_KEY_NAVIGATION))

        self.__grid.ShowRowLabels(False)
        self.__grid.ShowColLabels(True)

        self.__sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.__sizer.Add(self.__grid, flag=wx.EXPAND, proportion=1)

        self.SetSizer(self.__sizer)

        self.__grid.Bind(widgetgrid.EVT_WG_SELECT, self.__onGridSelect)

        lut.register(self.name, self.__lutChanged, 'added')
        lut.register(self.name, self.__lutChanged, 'removed')
        lut.register(self.name, self.__lutChanged, 'label')

        self.__overlay   = None
        self.__volLabels = None


    def destroy(self):
        """Must be called when this ``ComponentGrid`` is no longer needed.
        De-registers various property listeners, and calls
        :meth:`.FSLeyesPanel.destroy`.
        """

        self.__lut.deregister(self.name, 'added')
        self.__lut.deregister(self.name, 'removed')
        self.__lut.deregister(self.name, 'label')
        self.__deregisterCurrentOverlay()

        self.__lut = None

        fslpanel.FSLeyesPanel.destroy(self)


    def setOverlay(self, overlay, volLabels, refreshGrid=True):
        """Sets the :class:`.Image` to display component labels for.
        The :class:`.WidgetGrid` is re-populated to display the
        component-label mappings contained in the
        :class:`.VolumeLabels` instance associated with the overlay.

        :arg refreshGrid: If ``True`` (the default), the ``WidgetGrid``
                          displaying component labels is refreshed. This
                          flag is used internally (see
                          :meth:`__overlayTypeChanged`).
        """

        self.__deregisterCurrentOverlay()
        self.__grid.ClearGrid()

        if not (isinstance(overlay, fslimage.Image) and
                len(overlay.shape) == 4):
            self.__grid.Refresh()
            return

        log.debug('Registering new overlay: {}'.format(overlay))

        self.__overlay   = overlay
        self.__volLabels = volLabels
        display          = self.displayCtx.getDisplay(overlay)
        opts             = display.opts

        volLabels.register(             self.name, self.__labelsChanged)
        opts     .addListener('volume', self.name, self.__volumeChanged)
        display  .addListener('overlayType',
                              self.name,
                              self.__overlayTypeChanged)

        # We refresh the component grid on idle, in
        # case multiple calls to setOverlay are made
        # in quick succession - only the most recent
        # request will be executed.
        def doRefreshGrid():

            # The overlay might have been cleared
            # by the time this function gets called
            if self.__overlay is None:
                self.__grid.Refresh()
                return

            ncomps = self.__volLabels.numComponents()

            self.__grid.SetGridSize(ncomps, 2, growCols=[1])

            self.__grid.SetColLabel(0, strings.labels[self, 'componentColumn'])
            self.__grid.SetColLabel(1, strings.labels[self, 'labelColumn'])

            self.__recreateTags()
            self.__volumeChanged()

        if refreshGrid:
            idle.idle(doRefreshGrid,
                      name='{}_doRefreshGrid'.format(self.name),
                      skipIfQueued=True)


    def refreshTags(self, comps=None):
        """Clears and refreshes the tags on every :class:`.TextTagPanel` in
        the grid.

        :arg comps: Components to refresh. If ``None``, the tags for all
                    components are refreshed.
        """

        overlay   = self.__overlay
        volLabels = self.__volLabels
        numComps  = volLabels.numComponents()

        if comps is None:
            comps = list(range(numComps))

        if len(comps) == 0:
            return

        log.debug('Refreshing tags for {} [{}, ..., {}]'.format(
            overlay, comps[0], comps[-1]))

        for row in comps:

            tags = self.__grid.GetWidget(row, 1)

            tags.ClearTags()

            for label in volLabels.getLabels(row):
                tags.AddTag(volLabels.getDisplayLabel(label))

        self.__grid.Layout()


    def __deregisterCurrentOverlay(self):
        """Called when the selected overlay changes. De-registers listeners
        associated with the previously selected overlay, if necessary.
        """

        if self.__overlay is None:
            return

        overlay          = self.__overlay
        volLabels        = self.__volLabels
        self.__overlay   = None
        self.__volLabels = None

        volLabels.deregister(self.name)

        try:
            display = self.displayCtx.getDisplay(overlay)
            opts    = display.opts
            opts   .removeListener('volume',      self.name)
            display.removeListener('overlayType', self.name)

        except fsldisplay.InvalidOverlayError:
            pass


    def __overlayTypeChanged(self, *a):
        """Called when the :attr:`.Display.overlayType` of the currently
        displayed overlay changes. When the type of an overlay changes,
        a new :class:`.DisplayOpts` instance is created, so we need to
        re-register various property listeners with this new
        ``DisplayOpts`` instance.
        """
        self.setOverlay(self.__overlay, refreshGrid=False)


    def __recreateTags(self):
        """Called by :meth:`setOverlay`. Re-creates a :class:`.TextTagPanel`
        for every component in the :class:`.Image`.
        """

        volLabels = self.__volLabels
        numComps  = volLabels.numComponents()

        for i in range(numComps):

            tags = texttag.TextTagPanel(self.__grid,
                                        style=(texttag.TTP_ALLOW_NEW_TAGS |
                                               texttag.TTP_NO_DUPLICATES  |
                                               texttag.TTP_KEYBOARD_NAV))

            # Store the component number on the tag
            # panel, so we know which component we
            # are dealing with in the __onTagAdded
            # and __onTagRemoved methods.
            tags._componentIndex = i

            self.__grid.SetText(  i, 0, str(i + 1))
            self.__grid.SetWidget(i, 1, tags)

            tags.Bind(texttag.EVT_TTP_TAG_ADDED,   self.__onTagAdded)
            tags.Bind(texttag.EVT_TTP_TAG_REMOVED, self.__onTagRemoved)

        self.__grid.Refresh()
        self.__refreshTagOptions()
        self.refreshTags()
        self.Layout()


    def __refreshTagOptions(self):
        """Updates the options available on each :class:`.TextTagPanel`, from
        the entries in the melodic classification :class:`.LookupTable`.
        """

        overlay   = self.__overlay
        volLabels = self.__volLabels
        numComps  = volLabels.numComponents()

        log.debug('Updating component tag options for {}'.format(overlay))

        lut     = self.__lut
        labels  = [l.name   for l in lut]
        colours = [l.colour for l in lut]

        for i in range(len(colours)):
            colours[i] = [int(round(c * 255)) for c in colours[i]]

        for comp in range(numComps):
            tags = self.__grid.GetWidget(comp, 1)
            tags.SetOptions(labels, colours)


    def __onTagAdded(self, ev):
        """Called when a tag is added to a :class:`.TextTagPanel`. Adds the
        corresponding component-label mapping to the
        :class:`.VolumeLabels` instance.
        """

        tags      = ev.GetEventObject()
        label     = ev.tag
        comp      = tags._componentIndex
        lut       = self.__lut
        volLabels = self.__volLabels

        log.debug('Label added to component {} ("{}")'.format(comp, label))

        # Add the new label to the component
        with volLabels.skip(self.name):

            volLabels.addLabel(comp, label)

            # If the tag panel previously just contained
            # the 'Unknown' tag, remove that tag
            if tags.TagCount() == 2   and \
               tags.HasTag('Unknown') and \
               label.lower() != 'unknown':

                log.debug('Removing "unknown" tag from '
                          'component {}'.format(comp))

                volLabels.removeLabel(comp, 'Unknown')
                tags.RemoveTag('Unknown')

        # If the newly added tag is not in
        # the lookup table, add it in
        if lut.getByName(label) is None:
            colour = tags.GetTagColour(label)
            colour = [c / 255.0 for c in colour]

            log.debug('Adding new lookup table '
                      'entry for label {}'.format(label))

            with lut.skip(self.name, ('added', 'removed', 'label')):
                lut.new(name=label, colour=colour)

            self.__refreshTagOptions()
        self.__grid.Layout()


    def __onTagRemoved(self, ev):
        """Called when a tag is removed from a :class:`.TextTagPanel`.
        Removes the corresponding component-label mapping from the
        :class:`.VolumeLabels` instance.
        """

        tags      = ev.GetEventObject()
        label     = ev.tag
        comp      = tags._componentIndex
        volLabels = self.__volLabels

        log.debug('Label removed from component {} ("{}")'.format(comp, label))

        # Remove the label from
        # the melodic component
        with volLabels.skip(self.name):

            volLabels.removeLabel(comp, label)

            # If the tag panel now has no tags,
            # add the 'Unknown' tag back in.
            if len(volLabels.getLabels(comp)) == 0:

                log.debug('Adding "unknown" tag to '
                          'component {}'.format(comp))

                volLabels.addLabel(comp, 'Unknown')
                tags.AddTag('Unknown')

        self.__grid.FitInside()


    def __onGridSelect(self, ev):
        """Called when a row is selected on the :class:`.WidgetGrid`. Makes
        sure that the 'new tag' control in the corresponding
        :class:`.TextTagPanel` is focused.
        """

        component = ev.row
        opts      = self.displayCtx.getOpts(self.__overlay)

        log.debug('Grid row selected (component {}) - updating '
                  'overlay volume'.format(component))

        with props.skip(opts, 'volume', self.name):
            opts.volume = component

        tags = self.__grid.GetWidget(ev.row, 1)
        tags.FocusNewTagCtrl()


    def __volumeChanged(self, *a):
        """Called when the :attr:`.NiftiOpts.volume` property changes. Selects
        the corresponding row in the :class:`.WidgetGrid`.
        """

        # Only change the row if we are
        # currently visible, otherwise
        # this will screw up the focus.
        if not self.IsShown():
            return

        grid = self.__grid
        opts = self.displayCtx.getOpts(self.__overlay)

        log.debug('Overlay volume changed ({}) - updating '
                  'selected component'.format(opts.volume))

        # The setOverlay method updates the grid size on
        # the idle loop when the selected overlay changes,
        # so we have to update the selection on idle too,
        # otherwise the following sequence of events:
        #
        #  1. Overlay change (asynchronously schedules 3)
        #  2. Volume change (directly calls SetSelection on
        #     wrongly-sized grid)
        #  3. Grid refresh
        #
        # may raise an error
        idle.idle(grid.SetSelection, opts.volume, -1)


    def __labelsChanged(self, volLabels, topic, components):
        """Called on :class:`.VolumeLabels` notifications.
        Re-generates the tags shown on every :class:`.TextTagPanel`.
        """

        log.debug('Volume labels changed - '
                  'refreshing component grid tags')

        # The MelodicClassification
        # passes (component, label)
        # tuples, but we only care
        # about the components
        components = [c[0] for c in components]
        self.refreshTags(components)


    def __lutChanged(self, *a):
        """Called when the :attr:`.LookupTable.labels` change.
        Updates the options on every :class:`.TextTagPanel`.
        """
        log.debug('Lookup table changed - refreshing '
                  'component grid tag options')

        self.__refreshTagOptions()



#!/usr/bin/env python
#
# labelgrid.py - the LabelGrid class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`LabelGrid` class, which is used
by the :class:`.MelodicClassificationPanel`.
"""


import logging
import collections

import wx

import fsl.data.image             as fslimage
import fsl.utils.idle             as idle

import fsleyes_widgets.widgetgrid as widgetgrid
import fsleyes_widgets.texttag    as texttag

import fsleyes.panel              as fslpanel
import fsleyes.strings            as strings


log = logging.getLogger(__name__)


class LabelGrid(fslpanel.FSLeyesPanel):
    """The ``LabelGrid`` class is the inverse of the :class:`ComponentGrid`.
    It uses a :class:`.WidgetGrid` to display the label-component mappings
    present on the :class:`.VolumeLabels` instance associated with
    an :class:`.Image`. The ``Image`` and ``VolumeLabels`` instances are
    specified via the :meth:`setOverlay` method.

    The grid contains one row for each label, and a :class:`.TextTagPanel` is
    used to display the components associated with each label. Each
    ``TextTagPanel`` allows the user to add and remove components to/from the
    corresponding label.
    """


    def __init__(self, parent, overlayList, displayCtx, frame, lut):
        """Create a ``LabelGrid``.

        :arg parent:      The ``wx`` parent object.
        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg frame:       The :class:`.FSLeyesFrame` instance.
        :arg lut:         The :class:`.LookupTable` to be used to colour
                          component tags.
        """

        fslpanel.FSLeyesPanel.__init__(
            self, parent, overlayList, displayCtx, frame)

        self.__lut  = lut
        self.__grid = widgetgrid.WidgetGrid(
            self,
            style=(wx.VSCROLL                    |
                   widgetgrid.WG_SELECTABLE_ROWS |
                   widgetgrid.WG_KEY_NAVIGATION))

        # The LabelGrid displays one TextTagPanel
        # for each label that is currently displayed,
        # as:
        #
        #   { label_name : TextTagPanel }
        #
        # mappings.
        self.__labelTags = collections.OrderedDict()

        self.__sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.__sizer.Add(self.__grid, flag=wx.EXPAND, proportion=1)

        self.SetSizer(self.__sizer)

        self.__grid.Bind(widgetgrid.EVT_WG_SELECT, self.__onGridSelect)

        lut.register(self.name, self.__lutChanged, 'added')
        lut.register(self.name, self.__lutChanged, 'removed')
        lut.register(self.name, self.__lutChanged, 'label')

        self.__overlay = None


    def destroy(self):
        """Must be called when this ``LabelGrid`` is no longer needed.
        De-registers various property listeners, and calls
        :meth:`.FSLeyesPanel.destroy`.
        """

        self.__lut.deregister(self.name, 'added')
        self.__lut.deregister(self.name, 'removed')
        self.__lut.deregister(self.name, 'label')
        self.__lut = None

        self.__deregisterCurrentOverlay()

        fslpanel.FSLeyesPanel.destroy(self)


    def setOverlay(self, overlay, volLabels):
        """Set the :class:`.Image` shown on this ``LabelGrid``. A listener is
        registered with its :class:`.VolumeLabels`, and its component-label
        mappings displayed on the :class:`.WidgetGrid`.
        """

        self.__deregisterCurrentOverlay()
        self.__grid.ClearGrid()
        self.__labelTags.clear()

        if not (isinstance(overlay, fslimage.Image) and
                len(overlay.shape) == 4):
            self.__grid.Refresh()
            return

        log.debug('Registering new overlay: {}'.format(overlay))

        self.__overlay   = overlay
        self.__volLabels = volLabels

        volLabels.register(self.name, self.__labelsChanged)

        # We refresh the label grid on idle, in
        # case multiple calls to setOverlay are
        # made in quick succession - only the most
        # recent request will be executed.
        def createGrid():

            # The grid is initialised with length 0.
            # Rows for each label are are added in
            # the __createTags method, which is called
            # here, and as needed when the LUT or
            # MelodicClassification for the currently
            # selected overlay change.
            self.__grid.SetGridSize(0, 2, [1])
            self.__grid.ShowRowLabels(False)
            self.__grid.ShowColLabels(True)
            self.__grid.SetColLabel(0, strings.labels[self, 'labelColumn'])
            self.__grid.SetColLabel(1, strings.labels[self, 'componentColumn'])

            # The overlay might have been cleared
            # by the time this function gets called
            if self.__overlay is None:
                self.__grid.Refresh()
                return

            self.__createTags()
            self.refreshTags()
            self.__grid.Refresh()

        idle.idle(createGrid,
                  name='{}_createGrid'.format(self.name),
                  skipIfQueued=True)


    def __deregisterCurrentOverlay(self):
        """Called when the selected overlay changes. De-registers property
        listeners associated with the previously selected overlay, if
        necessary.
        """

        if self.__overlay is None:
            return

        volLabels        = self.__volLabels
        self.__overlay   = None
        self.__volLabels = None

        volLabels.deregister(self.name)


    def __createTags(self, labels=None):
        """Makes sure that a :class:`.TextTagPanel` exists for every label
        in the :class:`.LookupTable` and in the :class:`.VolumeLabels`
        for the current overlay.

        :arg labels: If ``None``, this method does what is described above.
                     Otherwise, this must be a list of tuples of
                     ``(name, display name)`` specifying labels which are
                     known not to be present, and for which a ``TextTagPanel``
                     needs to be created.

        :returns: ``True`` if one or more new ``TextTagPanel`` widgets was
                  created (and added to the :class:`.WidgetGrid`), ``False``
                  otherwise.
        """

        # This method should never be called
        # if the current overlay is not set.
        if self.__overlay is None:
            return

        if labels is None:

            volLabels = self.__volLabels
            allLabels = volLabels.getAllLabels()

            # We need to display a row for all
            # the labels from the lut, and all
            # classification labels which are
            # not in the lut.
            lutLabels = [(l.internalName, l.name)
                         for l in self.__lut]
            allLabels = [(l, volLabels.getDisplayLabel(l))
                         for l in allLabels
                         if l not in lutLabels]

            labels = lutLabels + allLabels

        newCreated = False

        log.debug('Creating tag panels for: {}'.format([l[0] for l in labels]))

        # TODO Ensure that there is always
        #      an 'unknown' tag... ?
        for label, displayName in labels:

            # A tag panel already exists for
            # this label, and must therefore
            # already be present in the grid.
            if label in self.__labelTags:
                continue

            newCreated = True
            tags = texttag.TextTagPanel(self.__grid,
                                        style=(texttag.TTP_NO_DUPLICATES |
                                               texttag.TTP_KEYBOARD_NAV))

            # This panel is associated with
            # this label - this association
            # will never change, so we can
            # safely store the label on the
            # tag itself, making it easy to
            # figure out the label given a
            # tag object.
            tags._label = displayName

            row = self.__grid.GetGridSize()[0]
            self.__grid.InsertRow(row)
            self.__grid.SetText(  row, 0, displayName)
            self.__grid.SetWidget(row, 1, tags)

            self.__labelTags[label] = tags

            tags.Bind(texttag.EVT_TTP_TAG_ADDED,   self.__onTagAdded)
            tags.Bind(texttag.EVT_TTP_TAG_REMOVED, self.__onTagRemoved)
            tags.Bind(texttag.EVT_TTP_TAG_SELECT,  self.__onTagSelect)

        return newCreated


    def refreshTags(self, labels=None):
        """Makes sure that the tags shown on each :class:`.TextTagPanel`
        are consistent with respect to the current state of the
        :class:`.VolumeLabels`.

        :arg labels: Labels to refresh. If ``None``, the tags for every
                     displayed label is refreshed.
        """

        if labels is None:
            labels = list(self.__labelTags.keys())

        log.debug('Refreshing tags for {}'.format(labels))

        # This method should never be called
        # if the current overlay is not set.
        lut       = self.__lut
        volLabels = self.__volLabels
        numComps  = volLabels.numComponents()
        compStrs  = [str(i) for i in range(1, numComps + 1)]

        for label in labels:

            tags     = self.__labelTags[label]
            lutLabel = lut.getByName(label)

            comps = volLabels.getComponents(label)

            # This is a label which is not in the
            # LUT. If we set colour to None, the
            # TextTagPanel will use a random colour.
            if lutLabel is None:
                colour = None

            # If this label is in the LUT, use
            # the colour associated with it.
            else:
                colour = lutLabel.colour
                colour = [int(round(c  * 255.0)) for c in colour]

            tags.ClearTags()
            tags.SetOptions(compStrs, [colour] * len(compStrs))

            for comp in comps:
                tags.AddTag(str(comp + 1))


    def __onGridSelect(self, ev):
        """Called when a row is selected in the :class:`.WidgetGrid`. Makes
        sure that  the first tag in the :class:`.TextTagPanel` has the focus.
        """

        tags = self.__grid.GetWidget(ev.row, 1)

        log.debug('Grid row selected (label "{}")'.format(tags._label))

        tags.FocusNewTagCtrl()


    def __onTagAdded(self, ev):
        """Called when a tag is added to a :class:`.TextTagPanel`. Adds
        the corresponding label-component mapping to the
        :class:`.VolumeLabels` instance.
        """

        tags      = ev.GetEventObject()
        volLabels = self.__volLabels
        comp      = int(ev.tag) - 1
        label     = tags._label

        log.debug('Component {} added to label {}'.format(comp, label))

        with volLabels.skip(self.name):

            # If this component now has two
            # labels, and the other label is
            # 'Unknown', remove the 'Unknown'
            # label.
            if len(volLabels.getLabels(comp)) == 1 and \
               label != 'unknown'                  and \
               volLabels.hasLabel(comp, 'unknown'):

                log.debug('Removing component {} from '
                          '"unknown" tag'.format(comp))

                volLabels.removeLabel(comp, 'unknown')
                self.__labelTags['unknown'].RemoveTag(str(comp + 1))

            volLabels.addComponent(label, comp)

        # The WidgetGrid doesn't
        # resize itself when its
        # contents change size
        self.__grid.Layout()


    def __onTagRemoved(self, ev):
        """Called when a tag is removed from a :class:`.TextTagPanel`. Removes
        the corresponding label-component mapping from the
        :class:`.VolumeLabels` instance.
        """

        tags      = ev.GetEventObject()
        volLabels = self.__volLabels
        lut       = self.__lut
        comp      = int(ev.tag) - 1
        label     = tags._label

        log.debug('Component {} removed from label {}'.format(comp, label))

        with volLabels.skip(self.name):

            volLabels.removeComponent(label, comp)

            # If the component has no more labels,
            # give it an 'Unknown' label
            if len(volLabels.getLabels(comp)) == 0:

                # What if there is no 'unknown'
                # entry in labelTags? Add it?
                label = lut.getByName('unknown')
                tags  = self.__labelTags['unknown']

                if label is None: name = volLabels.getDisplayLabel('unknown')
                else:             name = label.name

                # There should always be an 'unknown'
                # label in the LUT, but just in case.
                if label is None:
                    colour = None
                else:
                    colour = [int(round(c * 255.0)) for c in label.colour]

                log.debug('Adding component {} to '
                          '"unknown" tag'.format(comp))

                volLabels.addLabel(comp, name)
                tags.AddTag(str(comp + 1), colour)

        # The WidgetGrid doesn't
        # resize itself when its
        # contents change size
        self.__grid.Layout()


    def __onTagSelect(self, ev):
        """Called when a tag from a :class:`.TextTagPanel` is selected.
        Changes the current :attr:`.NiftiOpts.volume` to the component
        corresponding to the selected tag.
        """

        comp        = int(ev.tag) - 1
        overlay     = self.__overlay
        opts        = self.displayCtx.getOpts(overlay)

        log.debug('Component {} selected on label grid - '
                  'updating overlay volume'.format(comp))

        opts.volume = comp


    def __lutChanged(self, lut, topic, value):
        """Called when the :class:`.LookupTable` changes. Adds/removes/updates
        the displayed labels as needed.
        """

        if self.__overlay is None:
            return

        # The LookupTable passes us the LutLabel
        # object and its index in the LUT
        label, idx = value
        volLabels  = self.__volLabels

        # A label has been removed
        if topic == 'removed':

            # Delete the corresponding row
            # in the grid if there are no
            # components with the deleted
            # label.
            tags = self.__labelTags.get(label.internalName, None)

            if tags is not None and \
               len(volLabels.getComponents(label.internalName)) == 0:

                row = self.__grid.GetRow(tags)

                log.debug('LUT label {} removed - removing '
                          'corresponding row ({}) from widget '
                          'grid'.format(label.internalName, row))

                self.__labelTags.pop(label.internalName)
                self.__grid.DeleteRow(row)
                self.__grid.Layout()

        # A label has been added
        elif topic == 'added':

            # Are we already displaying a
            # label/tag which corresponds
            # to the new LUT label?
            tags = self.__labelTags.get(label.internalName, None)

            # If so, update its
            # colours, and return
            if tags is not None:
                colour = [int(round(c * 255.0)) for c in label.colour]

                for tag in tags.GetTags() + tags.GetOptions():
                    tags.SetTagColour(tag, colour)

            # Otherwise this is a new tag.
            # Create a tag/grid row for it
            else:
                self.__createTags([(label.internalName, label.name)])
                self.refreshTags(  [label.internalName])
                self.__grid.Refresh()

        # A property (name, colour) of an
        # existing label has changed.
        elif topic == 'label':

            # The only lut property that we track is
            # the label colour. If the label name
            # changes, we will ignore it; it will
            # ultimately be treated as a new label.
            tags = self.__labelTags.get(label.internalName, None)

            if tags is not None:
                colour = [int(round(c * 255.0)) for c in label.colour]

                for tag in tags.GetTags() + tags.GetOptions():
                    tags.SetTagColour(tag, colour)


    def __labelsChanged(self, volLabels, topic, components):
        """Called when the labels in the :class:`.VolumeLabels`
        associated with the current overlay change. Updates the displayed
        tags.
        """

        log.debug('Volume label changed - refreshing label grid tags')

        added = topic == 'added'

        for comp, label in components:
            tags = self.__labelTags.get(label, None)

            # Existing label
            if tags is not None:
                if added: tags.AddTag(   str(comp + 1))
                else:     tags.RemoveTag(str(comp + 1))

            # New label
            else:
                displayName = volLabels.getDisplayLabel(label)
                self.__createTags([(label, displayName)])
                self.refreshTags(  [label])
                self.__grid.Refresh()
